/*
 * NetOps.java
 *
 * Created on 31. bezen 2004, 17:05
 */

package SOFA.SOFAnet.Core;

import SOFA.SOFAnet.Repository.*;
import SOFA.SOFAnet.Transport.*;
import SOFA.SOFAnode.TR.Impl.*;
import SOFA.SOFAnode.TR.*;
import java.util.*;
import java.io.*;

/**
 * Network operations of the SOFAnet.
 *
 * @author  Ladislav Sobr
 */
public class NetOps
{
  private Repository rep;
  private TransportInterface transport;
  private LocalOps localOps;
  private ShareOps shareOps;
  
  private static class PushContractRule
  {
    public String nodeName;
    public String contractID;
    public String contractRule;
    public InputTrigger inputTrigger;
    
    public PushContractRule(String nodeName, String contractID, String contractRule, InputTrigger inputTrigger)
    {
      this.nodeName = nodeName;
      this.contractID = contractID;
      this.contractRule = contractRule;
      this.inputTrigger = inputTrigger;
    }
  }
    
  /** Creates a new instance of NetOps */
  public NetOps()
  {
  }
  
  public void setRepository(Repository repository)
  {
    rep = repository;
  }

  public void setTransportInterface(TransportInterface transport)
  {
    this.transport = transport;
  }
  
  public void setLocalOps(LocalOps localOps)
  {
    this.localOps = localOps;
  }

  public void setShareOps(ShareOps shareOps)
  {
    this.shareOps = shareOps;
  }

  
  /**
   * Tests whether node accepts push.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  offer
   * <li>  sourceNodeName   &nbsp;&nbsp;<i>    name of pushing node</i>
   * </ul>
   * <p>
   * Output: none
   *
   * @param params Input data for the call
   * @return true if push is acceptable
   */
  public boolean isPushAcceptable(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    boolean offer = params.isOffer();
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Push acceptance of bundle (or bundle offer) '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      return false;
    }
    
    InputTriggers inputTriggers = rep.getInputTriggers();
    
    synchronized (inputTriggers) {
    synchronized (inputTriggers.map)
    {
      Iterator it = inputTriggers.map.values().iterator();
      while (it.hasNext())
      {
        InputTrigger inputTrigger = (InputTrigger)it.next();

        int bundleType = inputTrigger.getBundleType();
        
        if (!inputTrigger.isPushSource()) continue;
        if (offer && bundleType != InputTrigger.BUNDLETYPE_OFFER && bundleType != InputTrigger.BUNDLETYPE_ALL) continue;
        if (!offer && bundleType != InputTrigger.BUNDLETYPE_FULL && bundleType != InputTrigger.BUNDLETYPE_ALL) continue;

        if (!inputTrigger.isValid()) continue;

        BundleNameFilter bundleFilter = inputTrigger.getBundleFilter();
        NodeNameFilter nodeFilter = inputTrigger.getNodeFilter();
        
        if (bundleFilter != null && !bundleFilter.pass(bundleInfo) ||
            nodeFilter != null && !nodeFilter.pass(nodeInfo)) continue;
        
        return true;
      }
    } //inputTriggers.map
    } //inputTriggers
    
    return false;
  }
  
  /**
   * Stores bundle/bundle offer/licence in the repository and preforms input triggers.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  bundleData
   * <li>  offer
   * <li>  sourceNodeName    &nbsp;&nbsp;<i>        name of pushing node</i>
   * <li>  licence           &nbsp;&nbsp;<i>        can be null</i>
   * </ul>
   * <p>
   * Output: none
   *
   * @param params Input data for the call
   */
  public void deliverBundle(IOParams params, boolean push) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    BundleData bundleData = params.getBundleData();
    Licence licence = params.getLicence();
    boolean offer = params.isOffer();

    if (offer) Reporter.startInfo("Delivering bundle offer '" + bundleName + " from node '" + nodeName + "' to local node");
    else Reporter.startInfo("Delivering bundle '" + bundleName + " from node '" + nodeName + "' to local node");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Delivery of bundle (or bundle offer) '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BundleOffers bundleOffers = rep.getBundleOffers();

    if (!offer)
    {
      synchronized (binBundles) {
      synchronized (licences)
      {
        BundleInfo binBundle = binBundles.getBinBundle(bundleName);
        LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
        Licence lic = licences.getLicence(bundleName);

        if (binBundle != null || lic != null || localInfo != null)
        {
          throw new CoreException("Delivery of bundle '" + bundleName + "' failed: Bundle is already present");
        }

        binBundle = binBundles.addBinBundle(bundleName);

        if (binBundle == null)
        {
          throw new CoreException("Delivery of bundle '" + bundleName + "' failed: Cannot add bundle to binary bundles");
        }

        if (!bundleData.copyToFile(binBundle.getFile(), true))
        {
          bundleData.delete();
          binBundles.deleteBinBundle(bundleName);
          throw new CoreException("Delivery of bundle '" + bundleName + "' failed: Error occured while copying bundle to binary bundles");
        }

        bundleData.delete();
        bundleData = null;

        if (licence != null && !licence.isEmpty())
        {
          lic = licences.addLicence(bundleName, licence.getType(), licence.getNumberOfCopies(),
                                    licence.getValidFrom(), licence.getValidTo(), licence.getText());
          
          if (lic == null)
          {
            binBundles.deleteBinBundle(bundleName);
            throw new CoreException("Delivery of bundle '" + bundleName + "' failed: Cannot add licence for bundle");
          }
        }
      } //licences
      } //binBundles
    
    }
    else
    {
      //bundle offer
      
      BundleImpl bundleOff = new BundleImpl();
      try
      {
        bundleOff._read(new FileInputStream(bundleData.getFile()));
      }
      catch (IOException e)
      {
        throw new CoreException("Delivery of bundle offer '" + bundleName + "' failed while decompressing bundle", e);
      }
      
      String sourceNodeName = "";
      ComponentInfo[] components = bundleOff.getComponents();
      
      if (components.length == 0)
      {
        throw new CoreException("Delivery of bundle offer '" + bundleName + "' failed: Empty bundle offer");
      }
      
      ComponentInfoImpl[] myComponents = new ComponentInfoImpl[components.length];
      for (int i = 0; i < components.length; i++)
      {
        if (i == 0)
        {
          sourceNodeName = components[i].getLocation();
          if (!sourceNodeName.startsWith("sofa://"))
          {
            bundleOff = null;
            throw new CoreException("Delivery of bundle offer '" + bundleName + "' failed: Invalid name of source node in bundle offer or not an bundle offer");
          }
        }
        else
        {
          if (sourceNodeName.compareTo(components[i].getLocation()) != 0)
          {
            bundleOff = null;
            throw new CoreException("Delivery of bundle offer '" + bundleName + "' failed: Inconsistent names of source node in bundle offer");
          }
        }
        myComponents[i] = new ComponentInfoImpl(components[i].getName(), components[i].getImplementationVersion());
      }
      
      bundleOff = null;

      BundleOffer bundleOffer = bundleOffers.addBundleOffer(bundleName, sourceNodeName, myComponents);
      
      if (bundleOffer == null)
      {
        throw new CoreException("Delivery of bundle offer '" + bundleName + "' failed: Cannot add bundle offer");
      }
    }

    //delete temporary file with bundle
    if (bundleData != null)
    {
      bundleData.delete();
      bundleData = null;
    }
    
    //perform input triggers
    int source;
    if (push) source = InputTrigger.SOURCE_PUSH;
    else  source = InputTrigger.SOURCE_PULL;
    performInputTriggers(bundleName, nodeName, source, offer);

    if (offer) Reporter.stopInfo("Bundle offer '" + bundleName + " from node '" + nodeName + "' delivered to local node");
    else Reporter.stopInfo("Bundle '" + bundleName + " from node '" + nodeName + "' delivered to local node");
  }

  /**
   * Wrapper for deliverBundle in push mode.
   */
  public void deliverBundlePush(IOParams params) throws CoreException
  {
    deliverBundle(params, true);
  }
  
  /**
   * Wrapper for deliverBundle in pull mode.
   */
  public void deliverBundlePull(IOParams params) throws CoreException
  {
    deliverBundle(params, false);
  }
  
  /**
   * Tests whether node accepts pull.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName  &nbsp;&nbsp;<i>     name of pulling node</i>
   * <li>  contractID      &nbsp;&nbsp;<i>     if output trigger has contract than contract's ID must match contractID to consider the output trigger</i>
   * </ul>
   * <p>
   * Output: none
   *
   * @param params Input data for the call
   * @return true if pull is acceptable
   */
  public boolean isPullAcceptable(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    String contractID = params.getContractID();

    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Pull acceptance of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      return false;
    }
    

    BinBundles binBundles = rep.getBinBundles();
    OutputTriggers outputTriggers = rep.getOutputTriggers();
    Contracts contracts = rep.getContracts();
    Licences licences = rep.getLicences();

    //test presence of binary bundle
    BundleInfo binBundle = binBundles.getBinBundle(bundleName);
    if (binBundle == null) return false;
    binBundle = null;

    synchronized (outputTriggers) {
    synchronized (outputTriggers.map)
    {
      Iterator it = outputTriggers.map.values().iterator();
      while (it.hasNext())
      {
        OutputTrigger outputTrigger = (OutputTrigger)it.next();

        if (!outputTrigger.isValid()) continue;

        BundleNameFilter bundleFilter = outputTrigger.getBundleFilter();
        NodeNameFilter nodeFilter = outputTrigger.getNodeFilter();

        if (bundleFilter != null && !bundleFilter.pass(bundleInfo) ||
            nodeFilter != null && !nodeFilter.pass(nodeInfo)) continue;

        if (outputTrigger.isContract())
        {
          if (outputTrigger.getContractID().compareTo(contractID) != 0) continue;
          if (contracts.passPull(bundleInfo, nodeInfo, outputTrigger.getContractID(), outputTrigger.getContractRule()) == null) continue;
        }
        else
        {
          synchronized (licences)
          {
            Licence lic = licences.getLicence(bundleName);
            if (lic != null && lic.withCopies() && lic.getNumberOfCopies() == 0) continue;
          } //licences
        }

        return true;
      }
    } //outputTriggers.map
    } //outputTriggers

    return false;
  }

  /**
   * Pull bundle from repository.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName  &nbsp;&nbsp;<i>     name of pulling node</i>
   * <li>  contractID      &nbsp;&nbsp;<i>     if output trigger has contract than contract's ID must match contractID to consider the output trigger</i>
   * </ul>
   * <p>
   * Output (on OK):
   * <ul>
   * <li>  bundleName
   * <li>  bundleData
   * <li>  licence
   * <li>  errCode == 0
   * </ul>
   * <p>
   * Output (on error):
   * <ul>
   * <li>  errCode
   *   <ul>
   *   <li>     1 == Cannot find binary bundle
   *   <li>     2 == Cannot create BundleData for bundle
   *   <li>     3 == Pull not allowed
   *   </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void preparePulledBundle(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    String contractID = params.getContractID();

    Reporter.startInfo("Preparing bundle '" + bundleName + "' for pull by node '" + nodeName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Preparation of pulled bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new CoreException("Preparation of pulled bundle '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
    }
    
    BinBundles binBundles = rep.getBinBundles();
    OutputTriggers outputTriggers = rep.getOutputTriggers();
    Contracts contracts = rep.getContracts();
    Licences licences = rep.getLicences();

    synchronized (binBundles) {
    synchronized (outputTriggers) {
    synchronized (outputTriggers.map) {
    synchronized (contracts) {
    synchronized (licences)
    {
      //test presence of binary bundle
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle == null)
      {
        params.setErrCode(1);
        return;
      }          

      Iterator it = outputTriggers.map.values().iterator();
      while (it.hasNext())
      {
        OutputTrigger outputTrigger = (OutputTrigger)it.next();

        if (!outputTrigger.isValid()) continue;

        BundleNameFilter bundleFilter = outputTrigger.getBundleFilter();
        NodeNameFilter nodeFilter = outputTrigger.getNodeFilter();

        if (bundleFilter != null && !bundleFilter.pass(bundleInfo) ||
            nodeFilter != null && !nodeFilter.pass(nodeInfo)) continue;

        ContractRule contractRule = null;
        Licence lic = null;

        if (outputTrigger.isContract())
        {
          if (outputTrigger.getContractID().compareTo(contractID) != 0) continue;
          contractRule = contracts.passPull(bundleInfo, nodeInfo, outputTrigger.getContractID(), outputTrigger.getContractRule());
          if (contractRule == null) continue;
        }
        else
        {
          lic = licences.getLicence(bundleName);
          if (lic != null && lic.withCopies() && lic.getNumberOfCopies() == 0) continue;
        }

        //we found output trigger allowing pull
        BundleData bundleData = new BundleData(binBundle.getFile(), true);
        if (bundleData.getFile() == null)
        {
          params.setErrCode(2);
        }

        params.setBundleData(bundleData);

        Licence licence = null;

        if (outputTrigger.isContract())
        {
          licence = contractRule.performPull();
          
          //remove contract rule, contract and output trigger if requested
          if (contractRule.toBeDeleted())
          {
            Contract contract = contractRule.getParentContract();
            contract.getRules().remove(contractRule.getName());
            if (contract.toBeDeleted()) contracts.deleteContract(contract.getName());
            else contract.saveToStorage();
            
            if (outputTrigger.getAutoDelete()) outputTriggers.deleteOutputTrigger(outputTrigger.getName());
          }
        }
        else
        {
          if (lic != null)
          {
            licence = (Licence) lic.clone();

            if (lic.withCopies())
            {
              licence.setNumberOfCopies(1);
              lic.decreaseNumberOfCopies(1);
              lic.saveToStorage();
            }
          }
        }

        params.setLicence(licence);
        params.setErrCode(0);
        Reporter.stopInfo("Bundle '" + bundleName + "' pulled by node '" + nodeName + "'");
        return;
      }
    } //licences
    } //contracts
    } //outputTriggers.map
    } //outputTriggers
    } //binBundles

    params.setErrCode(3);
    return;
  }
  
  /**
   * Performs actions of input triggers of the bundle (or bundle offer) when it arrives.
   * <p>
   * Actions are described in the input triggers. It can be:
   * <ul>
   * <li>install
   * <li>share
   * <li>allow_pull
   * <li>push
   * <li>push_offer
   * <li>delete
   * </ul>
   * <p>
   * All matching input triggers are considered together - first we find out which actions shoud be performed
   * from all input triggers. After that we distribute available licences among actions. Than the actions are
   * performed.
   *
   * @param bundleName name of bundle (or bundle offer)
   * @param nodeName name of source node of the bundle (even if the bundle is created on the node: for source == InputTrigger.SOURCE_MADE)
   * @param source how the bundle appeared on the node (creted/pulled/pushed) - see InputTrigger.SOURCE_... constants
   * @param offer if true, bundle offer arrived not the classical bundle
   */
  public void performInputTriggers(String bundleName, String nodeName, int source, boolean offer) throws CoreException
  {
    boolean actionInstall = false;
    boolean actionShare = false;
    boolean actionAllowPull = false;
    boolean actionPush = false;
    boolean actionPushOffer = false;
    boolean actionDelete = false;

    Reporter.startInfo("Performing input triggers on bundle '" + bundleName + "' from node '" + nodeName + "'");
    
    int asNoLicences = 0;
    ShareGroups asShareGroups = null;
    NodeNameFilter asNodeFilter = null;
    boolean asEquality = false;
    List alpOutputTriggers = new LinkedList();
    NodeNameLicList apNoContract = new NodeNameLicList();
    List apContract = new LinkedList();
    NodeNameList apoNodeNameList = new NodeNameList();
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Performing input triggers on '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new CoreException("Performing input triggers on '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
    }
    
    BinBundles binBundles = rep.getBinBundles();
    InputTriggers inputTriggers = rep.getInputTriggers();
    OutputTriggers outputTriggers = rep.getOutputTriggers();
    Contracts contracts = rep.getContracts();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BundleOffers bundleOffers = rep.getBundleOffers();
    
    synchronized (inputTriggers) {
    synchronized (licences) {
    synchronized (inputTriggers.map)
    {
      Licence lic = licences.getLicence(bundleName);
      
      Iterator it = inputTriggers.map.values().iterator();
      while (it.hasNext())
      {
        InputTrigger inputTrigger = (InputTrigger)it.next();
        
        int src = inputTrigger.getSource();
        int bundleType = inputTrigger.getBundleType();
        
        if ((src & source) == 0) continue;
        if (offer && bundleType != InputTrigger.BUNDLETYPE_OFFER && bundleType != InputTrigger.BUNDLETYPE_ALL) continue;
        if (!offer && bundleType != InputTrigger.BUNDLETYPE_FULL && bundleType != InputTrigger.BUNDLETYPE_ALL) continue;
        
        if (!inputTrigger.isValid()) continue;

        BundleNameFilter bundleFilter = inputTrigger.getBundleFilter();
        NodeNameFilter nodeFilter = inputTrigger.getNodeFilter();
        
        if (bundleFilter != null && !bundleFilter.pass(bundleInfo) ||
            nodeFilter != null && !nodeFilter.pass(nodeInfo)) continue;
        
        if ((inputTrigger.getActions() & InputTrigger.ACTION_INSTALL) != 0 && !offer)
        {
          actionInstall = true;
        }
        
        if ((inputTrigger.getActions() & InputTrigger.ACTION_SHARE) != 0 && !offer)
        {
          actionShare = true;
          int asNoLic = inputTrigger.getAsNoLicences();
          if (asNoLic == Licence.ONE_IMPLICIT_LICENCE) asNoLicences++;
          else if (asNoLic == Licence.ALL_LICENCES) asNoLicences = Licence.ALL_LICENCES;
          else asNoLicences += asNoLic;
          ShareGroups sg = inputTrigger.getAsShareGroups();
          if (sg != null)
          {
            if (asShareGroups == null) asShareGroups = new ShareGroups();
            asShareGroups.add(sg);
          }
          NodeNameFilter nnf = inputTrigger.getAsNodeFilter();
          if (nnf != null)
          {
            if (asNodeFilter == null) asNodeFilter = new NodeNameFilter();
            asNodeFilter.add(nnf);
          }
          
          if (inputTrigger.getAsEquality()) asEquality = true;
        }
        
        if ((inputTrigger.getActions() & InputTrigger.ACTION_ALLOW_PULL) != 0 && !offer)
        {
          actionAllowPull = true;
          OutputTrigger outputTrigger = inputTrigger.getAlpOutputTrigger();
          if (outputTrigger != null) alpOutputTriggers.add(outputTrigger.clone());
        }
        
        if ((inputTrigger.getActions() & InputTrigger.ACTION_PUSH) != 0)
        {
          if (!offer)
          {
            actionPush = true;
            
            //we have standard binary bundle, not only bundle offer
            if (inputTrigger.isContract())
            {
              NodeNameLicList apNodeNameLicList = inputTrigger.getApNodeNameLicList();
              String contractID = inputTrigger.getContractID();
              String contractRule = inputTrigger.getContractRule();

              if (apNodeNameLicList != null)
              {
                List li = apNodeNameLicList.getList();
                synchronized (li)
                {
                  Iterator it2 = li.iterator();
                  while (it2.hasNext())
                  {
                    NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it2.next();
                    apContract.add(new PushContractRule(nlp.nodeName, contractID, contractRule, inputTrigger));
                  }
                }
              }
            }
            else
            {
              apNoContract.add(inputTrigger.getApNodeNameLicList());
            }
          }
          else
          {
            actionPushOffer = true;
            
            //we have bundle offer, so act like if in "action PUSH OFFER"
            
            NodeNameLicList apNodeNameLicList = inputTrigger.getApNodeNameLicList();
            if (apNodeNameLicList != null)
            {
              List li = apNodeNameLicList.getList();
              synchronized (li)
              {
                Iterator it2 = li.iterator();
                while (it2.hasNext())
                {
                  NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it2.next();
                  apoNodeNameList.add(nlp.nodeName);
                }
              }
            }
            
          }
        }
        
        if ((inputTrigger.getActions() & InputTrigger.ACTION_PUSH_OFFER) != 0 && !offer)
        {
          actionPushOffer = true;
          apoNodeNameList.add(inputTrigger.getApoNodeNameList());
        }

        if ((inputTrigger.getActions() & InputTrigger.ACTION_DELETE) != 0)
        {
          actionDelete = true;
        }
      }
      
      
      //count licence copies for the actions
      
      if (lic != null && lic.withCopies())
      {
        int remains = lic.getNumberOfCopies();
        
        //add licence "to sharing" if install should be done but sharing has 0 licence copies
        if (actionInstall && asNoLicences == 0) asNoLicences = 1;
        
        //change ONE_IMPLICIT_LICENCE to 1
        if (asNoLicences == Licence.ONE_IMPLICIT_LICENCE) asNoLicences = 1;

        Iterator it2;
        
        it2 = apNoContract.getList().iterator();
        while (it2.hasNext())
        {
          NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it2.next();
          if (nlp.numberOfLicences == Licence.ONE_IMPLICIT_LICENCE) nlp.numberOfLicences = 1;
        }
        
        //handle all licence copies with specific number
        if (asNoLicences != Licence.ALL_LICENCES)
        {
          if (asNoLicences <= remains) remains -= asNoLicences;
          else
          {
            asNoLicences = remains;
            remains = 0;
          }
        }
        
        it2 = apNoContract.getList().iterator();
        while (it2.hasNext())
        {
          NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it2.next();
          if (nlp.numberOfLicences != Licence.ALL_LICENCES)
          {
            if (nlp.numberOfLicences <= remains) remains -= nlp.numberOfLicences;
            else
            {
              nlp.numberOfLicences = remains;
              remains = 0;
            }
          }
        }
        
        //handle "all remaining copies" flag (Licence.ALL_LICENCES)
        if (asNoLicences == Licence.ALL_LICENCES)
        {
          asNoLicences = remains;
          remains = 0;
        }
        
        it2 = apNoContract.getList().iterator();
        while (it2.hasNext())
        {
          NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it2.next();
          if (nlp.numberOfLicences == Licence.ALL_LICENCES)
          {
            nlp.numberOfLicences = remains;
            remains = 0;
          }
        }
        
        //now all informations about licence copies should contain concrete number of licences
      }
      else
      {
        //-no restrictions on licences
        //-change ONE_IMPLICIT_LICENCE to ALL_LICENCES

        //add licence "to sharing" if install should be done but sharing has 0 licence copies
        if (actionInstall && asNoLicences == 0) asNoLicences = Licence.ALL_LICENCES;
        
        if (asNoLicences == Licence.ONE_IMPLICIT_LICENCE) asNoLicences = Licence.ALL_LICENCES;

        Iterator it2;
        
        it2 = apNoContract.getList().iterator();
        while (it2.hasNext())
        {
          NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it2.next();
          if (nlp.numberOfLicences == Licence.ONE_IMPLICIT_LICENCE) nlp.numberOfLicences = Licence.ALL_LICENCES;
        }
      }
    } //inputTriggers.map
    } //licences
    } //inputTriggers
    
    
    //perform install action
    if (actionInstall)
    {
      if (asNoLicences != 0)
      {
        try
        {
          localOps.installBundle(bundleName, true);
        }
        catch (CoreException e)
        {
          Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot install bundle", e);
        }
      }
      else Reporter.error("Performing input triggers on '" + bundleName + "': Cannot install bundle - not enough licence copies");
    }
    
    //perform share action
    if (actionShare)
    {
      if (asNoLicences != 0)
      {
        try
        {
          shareOps.shareBundle(bundleName, asNoLicences, asShareGroups, asNodeFilter, asEquality, true);
        }
        catch (CoreException e)
        {
          Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot share bundle", e);
        }
      }
      else Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot share bundle - not enough licence copies");
    }
    
    //perform allow_pull action
    if (actionAllowPull)
    {
      synchronized (outputTriggers)
      {
        Iterator it2 = alpOutputTriggers.iterator();
        while (it2.hasNext())
        {
          OutputTrigger outputTrigger = (OutputTrigger)it2.next();
          outputTriggers.addOutputTrigger("", true, outputTrigger.getDescription(), outputTrigger.getAutoDelete(),
              outputTrigger.getBundleFilter(), outputTrigger.getNodeFilter(), outputTrigger.getValidFrom(),
              outputTrigger.getValidTo(), outputTrigger.getContractID(), outputTrigger.getContractRule());
        }
      } //outputTriggers
    }

    //perform push action
    if (actionPush)
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);

      //get BundleData
      BundleInfo bundle = null;
      BundleData bundleData = null;
      synchronized (binBundles)
      {
        bundle = binBundles.getBinBundle(bundleName);
        
        if (bundle != null)
        {
          bundleData = new BundleData(bundle.getFile(), true);
          if (bundleData.getFile() != null)
          {
            ioParams.setBundleData(bundleData);
          }
          else
          {
            Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot create BundleData");
          }
        }
        else
        {
          Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot find binary bundle");
        }
      } //binBundles
      
      if (bundleData != null)
      {
        ioParams.setOffer(false);
        ioParams.setPriority(IOParams.PRIORITY_DISTRIBUTION);
        
        int numNodes = 0;
        
        
        //handle destination nodes with contracts
        synchronized (inputTriggers) {
        synchronized (contracts)
        {
          Iterator it = apContract.iterator();
          while (it.hasNext())
          {
            PushContractRule pcr = (PushContractRule)it.next();
            Licence licence = contracts.performPush(bundleName, pcr.nodeName, pcr.contractID, pcr.contractRule);
            if (licence != null)
            {
              if (numNodes == 0)
              {
                ioParams.setNodeName(pcr.nodeName);
                if (!licence.isEmpty()) ioParams.setLicence(licence);
                else ioParams.setLicence(null);
              }
              else
              {
                IOParams.OtherNode otherNode = new IOParams.OtherNode();
                otherNode.nodeName = pcr.nodeName;
                if (!licence.isEmpty()) otherNode.licence = licence;
                else otherNode.licence = null;
                ioParams.addOtherNode(otherNode);
              }
              numNodes++;
              
              //remove contract rule, contract and input trigger if requested
              ContractRule contractRule = contracts.getContractRule(pcr.contractID, pcr.contractRule);
              if (contractRule.toBeDeleted())
              {
                Contract contract = contractRule.getParentContract();
                contract.getRules().remove(contractRule.getName());
                if (contract.toBeDeleted()) contracts.deleteContract(contract.getName());
                else contract.saveToStorage();

                if (pcr.inputTrigger.getAutoDelete()) inputTriggers.deleteInputTrigger(pcr.inputTrigger.getName());
              }
            }
          }
          
        } //contracts
        } //inputTriggers

        boolean maybeDelete = false;
        boolean deleteBundle = false;
        
        //handle destination nodes without contracts
        synchronized (licences)
        {
          Licence lic = licences.getLicence(bundleName);
          
          Iterator it = apNoContract.getList().iterator();
          while (it.hasNext())
          {
            NodeNameLicList.NodeLicPair nlp = (NodeNameLicList.NodeLicPair)it.next();
            
            Licence licence = null;
            boolean push = false;

            //we know that nlp.numberOfLicences is either >= 0 or Licence.ALL_LICENCES
            if (nlp.numberOfLicences > 0)
            {
              if (lic != null)
              {
                if (lic.withCopies())
                {
                  int num = lic.getNumberOfCopies();
                  if (num < nlp.numberOfLicences) nlp.numberOfLicences = num;
                  if (num > 0)
                  {
                    licence = (Licence)lic.clone();
                    licence.setNumberOfCopies(nlp.numberOfLicences);
                    lic.decreaseNumberOfCopies(nlp.numberOfLicences);
                    lic.saveToStorage();
                    push = true;
                  }
                }
                else
                {
                  licence = (Licence)lic.clone();
                  licence.setNumberOfCopies(nlp.numberOfLicences);
                  push = true;
                }
              }
              else
              {
                licence = new Licence();
                licence.setNumberOfCopies(nlp.numberOfLicences);
                push = true;
              }
            }
            else if (nlp.numberOfLicences == Licence.ALL_LICENCES)
            {
              if (lic != null)
              {
                if (lic.withCopies())
                {
                  //this shouldn't happen
                  int num = lic.getNumberOfCopies();
                  if (num > 0)
                  {
                    licence = (Licence)lic.clone();
                    licence.setNumberOfCopies(num);
                    lic.decreaseNumberOfCopies(num);
                    lic.saveToStorage();
                    push = true;
                  }
                }
                else
                {
                  licence = (Licence)lic.clone();
                  push = true;
                }
              }
              else push = true;
            }
            
            if (push)
            {
              if (numNodes == 0)
              {
                ioParams.setNodeName(nlp.nodeName);
                ioParams.setLicence(licence);
              }
              else
              {
                IOParams.OtherNode otherNode = new IOParams.OtherNode();
                otherNode.nodeName = nlp.nodeName;
                otherNode.licence = licence;
                ioParams.addOtherNode(otherNode);
              }
              numNodes++;
            }            
          }
          
          if (lic != null && lic.getType() == Licence.TYPE_PHYSICAL && lic.getNumberOfCopies() == 0) maybeDelete = true;

        } //licences
        
        //test whether to delete binary bundle because of no licence
        if (maybeDelete)
        {
          synchronized (localInfos)
          {
            LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
            if (localInfo == null || !localInfo.isInstalled() && localInfo.getLicence().getNumberOfCopies() == 0) deleteBundle = true;
          } //localInfos
        }
        
        //perform push on transport layer
        if (numNodes > 0)
        {
          try
          {
            transport.push(ioParams);
          }
          catch (TransportException e)
          {
            Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot push bundle", e);
          }
        }
        
        bundleData = null;
        bundle = null;
        
        //we have to delete binary bundle because of lack of licences
        if (deleteBundle)
        {
          try
          {
            localOps.deleteBinaryBundle(bundleName);
          }
          catch (CoreException e)
          {
            Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot delete binary bundle", e);
          }
        }
        
      }
    } //actionPush
    
    //perform push_offer action
    if (actionPushOffer)
    {
      IOParams ioParams = new IOParams();
 
      ioParams.setBundleName(bundleName);
      
      //prepare BundleData
      BundleData bundleData = null;
     
      synchronized (binBundles) {
      synchronized (bundleOffers)
      {
        BundleOffer bundleOffer = bundleOffers.getBundleOffer(bundleName);
        
        if (bundleOffer != null)
        {
          bundleData = bundleOffer.generateBundleData();
          if (bundleData == null)
          {
            Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot create BundleData for bundle offer");
          }
          else ioParams.setBundleData(bundleData);
        }
        else
        {
          try
          {
            bundleData = generateBundleDataForBundleOfferFromBinBundle(bundleName);
          }
          catch (CoreException e)
          {
            Reporter.error("Performing input triggers on '" + bundleName + "' failed while creatind BundleData for bundle offer", e);
          }
          
          ioParams.setBundleData(bundleData);
        }
      } //binBundles
      } //bundleOffers
      
      if (bundleData != null)
      {
        bundleData = null;  //we don't need it any more
        
        ioParams.setOffer(true);
        ioParams.setPriority(IOParams.PRIORITY_DISTRIBUTION);
        
        //prepare list of destination nodes
        int numNodes = 0;
        Iterator it = apoNodeNameList.getList().iterator();
        while (it.hasNext())
        {
          String outNodeName = (String)it.next();
          
          if (numNodes == 0) ioParams.setNodeName(outNodeName);
          else
          {
            IOParams.OtherNode otherNode = new IOParams.OtherNode();
            otherNode.nodeName = outNodeName;
            ioParams.addOtherNode(otherNode);
          }
          numNodes++;
        }
        
        //perform push on transport layer
        if (numNodes > 0)
        {
          try
          {
            transport.push(ioParams);
          }
          catch (TransportException e)
          {
            Reporter.error("Performing input triggers on '" + bundleName + "' failed: Cannot push bundle offer", e);
          }
        }
      }
    } //actionPushOffer
    
    //perform delete action
    if (actionDelete)
    {
      synchronized (binBundles) {
      synchronized (localInfos) {
      synchronized (licences) {
      synchronized (bundleOffers)
      {
        binBundles.deleteBinBundle(bundleName);
        bundleOffers.deleteBundleOffer(bundleName);

        LocalInfo localInfo = localInfos.getLocalInfo(bundleName);

        if (localInfo != null)
        {
          Reporter.error("Performing input triggers on '" + bundleName + "' failed: Bundle cannot be completely deleted because it is installed and/or shared");
        }
        else
        {
          licences.deleteLicence(bundleName);
        }
      } //bundleOffers
      } //localInfos
      } //licences
      } //binBundles
    }
    
    Reporter.stopInfo("Input triggers on bundle '" + bundleName + "' from node '" + nodeName + "' performed");
  }

  /**
   * Creates bundle offer from binary bundle on the local node.
   * 
   * @param bundleName name of bundle
   * @param overwrite if true the potentially existing bundle offer of that name is deleted first; else if bundle offer of that name exist an exception is thrown
   */
  public void createBundleOfferFromBinBundle(String bundleName, boolean overwrite) throws CoreException
  {
    BinBundles binBundles = rep.getBinBundles();
    BundleOffers bundleOffers = rep.getBundleOffers();

    Reporter.startInfo("Creating bundle offer for bundle '" + bundleName + "'");
    
    synchronized (binBundles) {
    synchronized (bundleOffers)
    {
      BundleOffer bundleOffer = bundleOffers.getBundleOffer(bundleName);
      if (bundleOffer != null)
      {
        throw new CoreException("Creation of bundle offer from binary bundle '" + bundleName + "' failed: Bundle offer already exists");
      }
      
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle == null)
      {
        throw new CoreException("Creation of bundle offer from binary bundle '" + bundleName + "' failed: Cannot find binary bundle");
      }

      //decompress binary bundle to the temp and create BundleImpl structure
      BundleImpl bundle = new BundleImpl();
      try
      {
        bundle._read(new FileInputStream(binBundle.getFile()));
      }
      catch (IOException e)
      {
        new CoreException("Creation of bundle offer from binary bundle '" + bundleName + "' failed while decompressing bundle", e);
      }
      
      //if bundle offer exits, delete it
      if (bundleOffer != null) bundleOffers.deleteBundleOffer(bundleName);
      
      //add new bundle offer
      ComponentInfo[] components = bundle.getComponents();
      if (components.length == 0)
      {
        new CoreException("Creation of bundle offer from binary bundle '" + bundleName + "' failed: Empty bundle");
      }
      
      ComponentInfoImpl[] myComponents = new ComponentInfoImpl[components.length];
      for (int i = 0; i < components.length; i++)
      {
        myComponents[i] = new ComponentInfoImpl(components[i].getName(), components[i].getImplementationVersion());
      }
      
      String sourceNodeName = NodeInfo.getLocalNodeName();
      bundleOffers.addBundleOffer(bundleName, sourceNodeName, myComponents);
      
      bundle = null;

      Reporter.stopInfo("Bundle offer for bundle '" + bundleName + "' successfully created");
      
    } //binBundles
    } //bundleOffers
  }

  /**
   * Creates BundleData version of bundle offer that is generated from binary bundle (but not stored among bundle offers!).
   *
   * @param bundleName name of bundle
   * @return BundleData version of generated bundle offer
   */
  public BundleData generateBundleDataForBundleOfferFromBinBundle(String bundleName) throws CoreException
  {
    BinBundles binBundles = rep.getBinBundles();

    synchronized (binBundles)
    {
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle == null)
      {
        throw new CoreException("Creation of bundle offer file from binary bundle '" + bundleName + "' failed: Cannot find binary bundle");
      }

      //decompress binary bundle to the temp and create BundleImpl structure
      BundleImpl bundle = new BundleImpl();
      try
      {
        bundle._read(new FileInputStream(binBundle.getFile()));
      }
      catch (IOException e)
      {
        new CoreException("Creation of BundleData for bundle offer from binary bundle '" + bundleName + "' failed while decompressing bundle", e);
      }
      
      BundleImpl bundleOff = new BundleImpl();
      ComponentInfo[] components = bundle.getComponents();
      if (components.length == 0)
      {
        throw new CoreException("Creation of BundleData for bundle offer from binary bundle '" + bundleName + "' failed: Empty bundle");
      }
      
      for (int i = 0; i < components.length; i++)
      {
        bundleOff.add(new ComponentInfoImpl(components[i].getName(), components[i].getImplementationVersion()), BundleImpl.NO_FILES);
      }

      bundle = null;
      
      File tempFile = new File(BundleData.generateFilename());
      
      try
      {
        FileOutputStream os = new FileOutputStream(tempFile);
        bundleOff._write(os);
        os.close();
      }
      catch (Exception e)
      {
        throw new CoreException("Creation of BundleData for bundle offer from binary bundle '" + bundleName + "' failed while saving binary bundle to file", e);
      }
      
      bundleOff = null;

      return new BundleData(tempFile, false);
    } //binBundles
  }

  /**
   * Tests whether push of the bundle or bundle offer is accepted by destination node.
   * This is "client" of the "testPush" operation.
   *
   * @param bundleName name of bundle (or bundle offer)
   * @param nodeName name of destination node
   * @param offer if true push of bundle offer is tested; otherwise classical bundle is concerned
   * @return true if the push can be accepted by destination node
   */
  public boolean testPush(String bundleName, String nodeName, boolean offer) throws CoreException
  {
    IOParams ioParams = new IOParams();
    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    ioParams.setOffer(offer);
    
     //perform testPush on transport layer
    try
    {
      transport.testPush(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Test push of '" + bundleName + "' failed", e);
    }
    
    return ioParams.getErrCode() == 0;
  }
  
  /**
   * Pushes bundle to the destination node.
   * This is "client" of the "push" operation.
   *
   * @param bundleName name of bundle
   * @param nodeName name of destination node
   * @param numberOfLicences positive integer of Licence.ALL_LICENCES ; no effect if contract is present
   */
  public void pushBundle(String bundleName, String nodeName, int numberOfLicences) throws CoreException
  {
    Reporter.startInfo("Pushing bundle '" + bundleName + "' to node '" + nodeName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Push of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new CoreException("Push of bundle '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
    }
 
    IOParams ioParams = new IOParams();
    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    ioParams.setOffer(false);
 
    BinBundles binBundles = rep.getBinBundles();
    Contracts contracts = rep.getContracts();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();

    //get BundleData
    BundleInfo bundle = null;
    BundleData bundleData = null;
    synchronized (binBundles)
    {
      bundle = binBundles.getBinBundle(bundleName);

      if (bundle == null) throw new CoreException("Push of bundle '" + bundleName + "' failed: Cannot find binary bundle");

      bundleData = new BundleData(bundle.getFile(), true);
      if (bundleData.getFile() == null) throw new CoreException("Push of bundle '" + bundleName + "' failed: Cannot create BundleData");

      ioParams.setBundleData(bundleData);
    } //binBundles
    
    ioParams.setPriority(IOParams.PRIORITY_INTERACTIVE);

    Licence licence = null;
    boolean deleteBundle = false;
    
    //check whether contracts are involved
    synchronized (contracts)
    {
      ContractRule contractRule = contracts.passPush(bundleInfo, nodeInfo);
      if (contractRule != null) licence = contractRule.performPush(nodeInfo);
    } //contracts

    if (licence == null)
    {
      boolean maybeDelete = false;
      
      //try licence files
      synchronized (licences)
      {
        Licence lic = licences.getLicence(bundleName);
        
        if (lic != null)
        {
          if (lic.withCopies())
          {
            int licNum = lic.getNumberOfCopies();
            if (numberOfLicences == Licence.ALL_LICENCES) numberOfLicences = licNum;
            
            if (numberOfLicences > licNum)
            {
              throw new CoreException("Push of bundle '" + bundleName + "' failed: Not enough licence copies");
            }
            
            licence = (Licence)lic.clone();
            licence.setNumberOfCopies(numberOfLicences);
            lic.decreaseNumberOfCopies(numberOfLicences);
            lic.saveToStorage();
            
            if (lic.getType() == Licence.TYPE_PHYSICAL && lic.getNumberOfCopies() == 0) maybeDelete = true;
          }
          else
          {
            licence = (Licence)lic.clone();
            if (numberOfLicences >= 0) licence.setNumberOfCopies(numberOfLicences);
          }
        }
        else
        {
          if (numberOfLicences >= 0)
          {
            licence = new Licence();
            licence.setNumberOfCopies(numberOfLicences);
          }
        }
      } //licences
      
      //test whether to delete binary bundle because of no licence
      if (maybeDelete)
      {
        synchronized (localInfos)
        {
          LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
          if (localInfo == null || !localInfo.isInstalled() && localInfo.getLicence().getNumberOfCopies() == 0) deleteBundle = true;
        } //localInfos
      }
    }
    
    ioParams.setLicence(licence);
    
    //perform push on transport layer
    try
    {
      transport.push(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Push of bundle '" + bundleName + "' failed", e);
    }
    
    //we have to delete binary bundle because of lack of licences
    if (deleteBundle)
    {
      try
      {
        localOps.deleteBinaryBundle(bundleName);
      }
      catch (CoreException e)
      {
        throw new CoreException("Push of bundle '" + bundleName + "' failed: Cannot delete binary bundle", e);
      }
    }
    
    Reporter.stopInfo("Bundle '" + bundleName + "' scheduled to push to node '" + nodeName + "'");    
  }

  /**
   * Pushes bundle offer to the destination node. Bundle offer must already exist on the local node.
   * This is "client" of the "push" operation.
   *
   * @param bundleName name of bundle (or bundle offer)
   * @param nodeName name of destination node
   */
  public void pushBundleOffer(String bundleName, String nodeName) throws CoreException
  {
    Reporter.startInfo("Pushing bundle offer '" + bundleName + "' to node '" + nodeName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Push of bundle offer '" + bundleName + "' failed: Invalid name of bundle offer", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new CoreException("Push of bundle offer '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
    }
 
    IOParams ioParams = new IOParams();
    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    ioParams.setOffer(true);
  
    BundleData bundleData = null;
    BundleOffers bundleOffers = rep.getBundleOffers();
    
    synchronized (bundleOffers)
    {
      BundleOffer bundleOffer = bundleOffers.getBundleOffer(bundleName);
      if (bundleOffer == null) throw new CoreException("Push of bundle offer '" + bundleName + "' failed: Cannot find bundle offer");

      bundleData = bundleOffer.generateBundleData();
      if (bundleData == null) throw new CoreException("Push of bundle failed'" + bundleName + "' failed: Cannot create BundleData from bundle offer");
    } //bundleOffers

    ioParams.setPriority(IOParams.PRIORITY_INTERACTIVE);
    ioParams.setBundleData(bundleData);
    ioParams.setLicence(null);
    
    //perform push on transport layer
    try
    {
      transport.push(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Push of bundle offer'" + bundleName + "' failed", e);
    }
    
    Reporter.stopInfo("Bundle offer '" + bundleName + "' scheduled to push to node '" + nodeName + "'");

  }
  
  /**
   * Tests whether pull (download) of the bundle is accepted by source node.
   * This is "client" of the "testPull" operation.
   *
   * @param bundleName name of bundle
   * @param nodeName name of source node
   * @param contractID ID of the contract that we have at the source node ; use empty string for no contract
   * @return true if the pull can be accepted by the source node
   */
  public boolean testPull(String bundleName, String nodeName, String contractID) throws CoreException
  {
    IOParams ioParams = new IOParams();
    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    ioParams.setContractID(contractID);
    
    //perform testPull on transport layer
    try
    {
      transport.testPull(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Test pull of '" + bundleName + "' failed", e);
    }
    
    return ioParams.getErrCode() == 0;
  }
  
  
  /**
   * Pulls (downloads) bundle from the source node.
   * This is "client" of the "pull" operation.
   * <p>
   * The actions of input triggers are performed when the bundle arrives.
   * <p>
   * This "pull" is just an order to the transport layer to do that action. It needen't be performed immediatelly!
   *
   * @param bundleName name of bundle
   * @param nodeName name of source node
   * @param contractID ID of the contract that we have at the source node ; use empty string for no contract
   */
  public void pullBundle(String bundleName, String nodeName, String contractID) throws CoreException
  {
    Reporter.startInfo("Pulling bundle '" + bundleName + "' from node '" + nodeName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Pull of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new CoreException("Pull of bundle '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
    }
 
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    
    if (binBundles.getBinBundle(bundleName) != null ||
        licences.getLicence(bundleName) != null ||
        localInfos.getLocalInfo(bundleName) != null)
    {
      throw new CoreException("Pull of bundle '" + bundleName + "' failed: Bundle is already present (binary bundle/licence/local info)");
    }
        
    IOParams ioParams = new IOParams();
    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    ioParams.setContractID(contractID);
    ioParams.setPriority(IOParams.PRIORITY_INTERACTIVE);
   
    //perform testPull on transport layer
    try
    {
      transport.pull(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Pull of '" + bundleName + "' failed", e);
    }
    
    Reporter.stopInfo("Bundle '" + bundleName + "' scheduled to pull from node '" + nodeName + "'");    
  }
  
  /**
   * Pulls (downloads) bundle from the source node.
   * This is "client" of the "synchroPull" operation.
   * <p>
   * The actions of input triggers are performed when the bundle arrives.
   * <p>
   * This "pull" is performed immediatelly and it is finished before return of the function.
   *
   * @param bundleName name of bundle
   * @param nodeName name of source node
   * @param contractID ID of the contract that we have at the source node ; use empty string for no contract
   */
  public void synchroPullBundle(String bundleName, String nodeName, String contractID) throws CoreException
  {
    Reporter.startInfo("Synchonously pulling bundle '" + bundleName + "' from node '" + nodeName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Pull of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new CoreException("Pull of bundle '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
    }
 
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    
    if (binBundles.getBinBundle(bundleName) != null ||
        licences.getLicence(bundleName) != null ||
        localInfos.getLocalInfo(bundleName) != null)
    {
      throw new CoreException("Pull of bundle '" + bundleName + "' failed: Bundle is already present (binary bundle/licence/local info)");
    }
        
    IOParams ioParams = new IOParams();
    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    ioParams.setContractID(contractID);
    ioParams.setPriority(IOParams.PRIORITY_INTERACTIVE);
   
    //perform testPull on transport layer
    try
    {
      transport.synchroPull(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Pull of '" + bundleName + "' failed", e);
    }
    
    int errCode = ioParams.getErrCode();
    if (errCode != 0)
    {
      switch (errCode)
      {
      case 1: throw new CoreException("Pull of '" + bundleName + "' failed: Cannot find binary bundle on remote node");
      case 2: throw new CoreException("Pull of '" + bundleName + "' failed: Cannot create BundleData for bundle");
      case 3: throw new CoreException("Pull of '" + bundleName + "' failed: Pull not allowed by remote node");
      default: throw new CoreException("Pull of '" + bundleName + "' failed: Transport layer failed");
      }
    }
    
    ioParams.setBundleName(bundleName);
    ioParams.setSourceNodeName(nodeName);
    
    deliverBundlePull(ioParams);
    
    Reporter.stopInfo("Bundle '" + bundleName + "' synchronously pulled from node '" + nodeName + "'");    
  }
 
  /**
   * Generates input and output triggers that performs the rules introduced by the contract.
   *
   * @param contractID ID (name) of contract
   * @param firstDelete if true all input and output triggers that belongs to the contract are deleted first
   */
  public void generateTriggersOfContract(String contractID, boolean firstDelete) throws CoreException
  {
    Reporter.startInfo("Generating triggers of contract '" + contractID + "'");
    
    Contracts contracts = rep.getContracts();
    InputTriggers inputTriggers = rep.getInputTriggers();
    OutputTriggers outputTriggers = rep.getOutputTriggers();
    
    Contract contract = contracts.getContract(contractID);
    if (contract == null) throw new CoreException("Cannot find Contract '" + contractID + "'");
    
    contract.generateTriggers(firstDelete, inputTriggers, outputTriggers);    
    
    Reporter.stopInfo("Triggers of contract '" + contractID + "' generated");
  }
  
  /**
   * Deletes input and output triggers belongs to the contract
   *
   * @param contractID ID (name) of contract
   */
  public void deleteTriggersOfContract(String contractID) throws CoreException
  {
    Reporter.startInfo("Deleting triggers of contract '" + contractID + "'");
    
    Contracts contracts = rep.getContracts();
    InputTriggers inputTriggers = rep.getInputTriggers();
    OutputTriggers outputTriggers = rep.getOutputTriggers();
    
    Contract contract = contracts.getContract(contractID);
    if (contract == null) throw new CoreException("Cannot find Contract '" + contractID + "'");
    
    contract.deleteTriggers(inputTriggers, outputTriggers);    

    Reporter.stopInfo("Triggers of contract '" + contractID + "' deleted");
  }
}