/*
 * SearchOps.java
 *
 * Created on 28. duben 2004, 15:47
 */

package SOFA.SOFAnet.Core;

import SOFA.SOFAnet.Repository.*;
import SOFA.SOFAnet.Search.*;
import SOFA.SOFAnode.TR.ComponentInfo;
import SOFA.SOFAnode.TR.Impl.ComponentInfoImpl;
import SOFA.Util.VMProperties;
import java.util.*;

/**
 * Search operations of the SOFAnet.
 *
 * @author  Ladislav Sobr
 */
public class SearchOps
{
  private static int componentSearchTimeout;
  
  private final boolean shareAllComponents;
  private final String myNodeName;
  private Repository rep;
  private SearchInterface search;
  private LocalOps localOps;
  
  
  /** Creates a new instance of SearchOps */
  public SearchOps()
  {
    String shareAllComponentsStr = System.getProperty(VMProperties.NET_SHARE_ALL_COMPONENTS, VMProperties.NET_SHARE_ALL_COMPONENTS_DEF).trim().toLowerCase();
    shareAllComponents = shareAllComponentsStr.compareTo("1") == 0 || shareAllComponentsStr.compareTo("true") == 0;
    
    componentSearchTimeout = 5000;
    String componentSearchTimeoutStr = System.getProperty(VMProperties.NET_SEARCH_TIMEOUT, VMProperties.NET_SEARCH_TIMEOUT_DEF);
    
    try
    {
      componentSearchTimeout = Integer.parseInt(componentSearchTimeoutStr);
    }
    catch (NumberFormatException e)
    {
    }
    
    myNodeName = NodeInfo.getLocalNodeName();
  }
  
  public void setRepository(Repository repository)
  {
    rep = repository;
  }
  
  public void setSearchInterface(SearchInterface search)
  {
    this.search = search;
  }

  public void setLocalOps(LocalOps localOps)
  {
    this.localOps = localOps;
  }

  /**
   * Evaluates whether this node can satisfy the search request.
   * If so, the search reply is created and sended to the requesting node.
   *
   * @param requestID identification of the request (in the search network)
   * @param searchPattern describes, what is requested
   * @param firstOnly if true, at most one reply item is requested
   * @param shareGroup specification of the group that is been searched through; empty string if no group should be specified
   */
  public void evaluateRequest(SearchRequestID requestID, SearchPattern searchPattern, boolean firstOnly, String shareGroup)
  {
    boolean pattern = searchPattern.isPattern();
    boolean component = searchPattern.isComponent();
    int completeComponent = searchPattern.getCompleteComponent();
    String str = searchPattern.getStr();
    
    List searchReplies = new LinkedList();
    
    LocalInfos localInfos = rep.getLocalInfos();
    BundleContents sharedBundleContents = rep.getSharedBundleContents();
    CompBundleMap sharedCompBundleMap = sharedBundleContents.getCompBundleMap();
    
    synchronized (localInfos)
    {
      //try bundles first
      if (pattern)
      {
        synchronized (localInfos.map)
        {
          Iterator it = localInfos.map.entrySet().iterator();
          while (it.hasNext())
          {
            Map.Entry entry = (Map.Entry)it.next();
            String bundleName = (String)entry.getKey();
            LocalInfo localInfo = (LocalInfo)entry.getValue();
            ShareGroups shareGroups = localInfo.getSMShareGroups();
            if (localInfo.isShareManager() && 
                (shareGroups == null || shareGroup.length() == 0 || shareGroups.contains(shareGroup)))
            {
              try
              {
                if (BundleInfo.matchPattern(bundleName, str))
                {
                  searchReplies.add(new SearchReplyItem(myNodeName, bundleName));
                  if (firstOnly) break;
                }
              }
              catch (BundleInfo.InvalidBundleNameException e)
              {
              }
            }
          }
        }
      }
      else
      {
        //str contains exact name (of bundle or component)
        LocalInfo localInfo = localInfos.getLocalInfo(str);
        
        if (localInfo != null && localInfo.isShareManager())
        {
          ShareGroups shareGroups = localInfo.getSMShareGroups();
          if (shareGroups == null || shareGroup.length() == 0 || shareGroups.contains(shareGroup))
          {
            searchReplies.add(new SearchReplyItem(myNodeName, localInfo.getName()));
          }
        }
      }
    } //localInfos

    
    //test correctness of name of the component (if it should be component)
    ComponentInfo componentInfo = null;
    if (component && !pattern)
    {
      //test correctness of str as name of component
      BundleInfo bundleInfo = new BundleInfo();
      try
      {
        bundleInfo.fromBundleName(str);
      }
      catch (BundleInfo.InvalidBundleNameException e)
      {
        Reporter.error("Evaluation of search request (" + requestID.getRequestMark() + ",'" + requestID.getSourceNode() + "') failed: Invalid name of component: " + str, e);
        bundleInfo = null;
      }
      if (bundleInfo != null && !bundleInfo.isComponent()) bundleInfo = null;
      if (bundleInfo != null) componentInfo = new ComponentInfoImpl(bundleInfo.getShortName(), bundleInfo.getVersion());
    }
    
    if (!component || !pattern && componentInfo == null || firstOnly && !searchReplies.isEmpty())
    {
      if (searchReplies.isEmpty()) return;
      
      try
      {
        search.reply(new SearchReply(requestID, searchReplies));
      }
      catch (SearchException e)
      {
        Reporter.error("Evaluation of search request (" + requestID.getRequestMark() + ",'" + requestID.getSourceNode() + "') failed while replying", e);
      }
      return;
    }
    
    //no we know "str" is component name/pattern
    
    //go through contents of bundles (but not if completeComponent == CC_ONLY, because we cannot assure that the bundle contain complete component)
    if (completeComponent != SearchPattern.CC_ONLY)
    {
      if (pattern)
      {
        synchronized (localInfos) {
        synchronized (sharedBundleContents) {
        synchronized (sharedCompBundleMap.map)      //just to be sure
        {
          Iterator it = sharedCompBundleMap.map.entrySet().iterator();
          while (it.hasNext())
          {
            Map.Entry entry = (Map.Entry)it.next();
            String componentName = (String)entry.getKey();
            String bundleName = (String)entry.getValue();
            
            try
            {
              if (BundleInfo.matchPattern(componentName, str))
              {
                //ok, we found the component that matches search pattern, but still we have to find out, whether we are the share manager of the component
                LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
                if (localInfo != null && localInfo.isShareManager())
                {
                  searchReplies.add(new SearchReplyItem(myNodeName, bundleName));
                  if (firstOnly) break;
                }
              }
            }
            catch (BundleInfo.InvalidBundleNameException e)
            {
            }
          }
        
        } //sharedCompBundleMap.map  
        } //sharedBundleContents
        } //localInfos
      }
      else
      {
        synchronized (localInfos)
        {
          String bundleName = sharedCompBundleMap.findBundle(str);
          if (bundleName != null)
          {
            //ok, we found the component, but still we have to find out, whether we are the share manager of the component
            LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
            if (localInfo != null && localInfo.isShareManager())
            {
              searchReplies.add(new SearchReplyItem(myNodeName, bundleName));
            }
          }
        } //localInfos
      }
    }
    
    if (!shareAllComponents || firstOnly && !searchReplies.isEmpty())
    {
      if (searchReplies.isEmpty()) return;
      
      try
      {
        search.reply(new SearchReply(requestID, searchReplies));
      }
      catch (SearchException e)
      {
        Reporter.error("Evaluation of search request (" + requestID.getRequestMark() + ",'" + requestID.getSourceNode() + "') failed while replying", e);
      }
      return;
    }
    
    //now we know, that shareAllComponents == true - we will try to create component-bundle from TR
    
    boolean haveCompleteComponent = false;
    //try to find complete components (if we are requested to do it)
    if (completeComponent == SearchPattern.CC_TRY || completeComponent == SearchPattern.CC_ONLY)
    {
      if (pattern)
      {
        ComponentInfo[] comps = localOps.listComponentsPresentInTR();
        for (int i = 0; i < comps.length; i++)
        {
          if (localOps.testComponentPresenceInTR(comps[i], true))
          {
            searchReplies.add(new SearchReplyItem(myNodeName, BundleInfo.makeComponentBundleName(comps[i].getName(), comps[i].getImplementationVersion(), true)));
            haveCompleteComponent = true;
            if (firstOnly) break;
          }
        }
      }
      else
      {
        if (localOps.testComponentPresenceInTR(componentInfo, true))
        {
          searchReplies.add(new SearchReplyItem(myNodeName, BundleInfo.makeComponentBundleName(componentInfo.getName(), componentInfo.getImplementationVersion(), true)));
          haveCompleteComponent = true;
        }
      }
    }
    
    //try to find single components (if we are requested to do it)
    if (completeComponent == SearchPattern.CC_NO || completeComponent == SearchPattern.CC_TRY && !haveCompleteComponent)
    {
      if (pattern)
      {
        ComponentInfo[] comps = localOps.listComponentsPresentInTR();
        for (int i = 0; i < comps.length; i++)
        {
          if (localOps.testComponentPresenceInTR(comps[i], false))
          {
            searchReplies.add(new SearchReplyItem(myNodeName, BundleInfo.makeComponentBundleName(comps[i].getName(), comps[i].getImplementationVersion(), false)));
            if (firstOnly) break;
          }
        }
      }
      else
      {
        if (localOps.testComponentPresenceInTR(componentInfo, false))
        {
          searchReplies.add(new SearchReplyItem(myNodeName, BundleInfo.makeComponentBundleName(componentInfo.getName(), componentInfo.getImplementationVersion(), false)));
          haveCompleteComponent = true;
        }
      }
    }
    
    if (searchReplies.isEmpty()) return;

    try
    {
      search.reply(new SearchReply(requestID, searchReplies));
    }
    catch (SearchException e)
    {
      Reporter.error("Evaluation of search request (" + requestID.getRequestMark() + ",'" + requestID.getSourceNode() + "') failed while replying", e);
    }
  }
  
  /**
   * Searches for the node that can provide bundle containing specified component
   * 
   * @param componentFullName full name of the component in form name[version]
   * @return search reply item if successfull, null otherwise
   */
  public SearchReplyItem searchForBundleWithComponent(String componentFullName) throws CoreException
  {
    Reporter.startInfo("Searching for component '" + componentFullName + "'");

    SearchPattern pattern = new SearchPattern();
    pattern.setStr(componentFullName);
    pattern.setPattern(false);
    pattern.setComponent(true);
    pattern.setCompleteComponent(SearchPattern.CC_TRY);
    
    SearchRequest request = new SearchRequest();
    request.setSearchPattern(pattern);
    request.setFirstOnly(true);
    request.setShareGroup("");
    request.setTimeout(componentSearchTimeout);
    
    try
    {
      search.request(request);
    }
    catch (SearchException e)
    {
      throw new CoreException(e.getMessage(), e);
    }

    List replies = request.getReplies();
    if (request.getErrCode() == 0 && replies != null && replies.size() > 0)
    {
      SearchReplyItem replyItem = (SearchReplyItem)replies.get(0);
      Reporter.stopInfo("Component '" + componentFullName + "' found at '" + replyItem.getNodeName() + "' in bundle '" + replyItem.getBundleName() + "'");
      return replyItem;
    }
    else
    {
      Reporter.stopInfo("Component '" + componentFullName + "' not found");
      return null;
    }
  }

  /**
   * Searches for the nodes and bundles that can satisfy request.
   * 
   * @param searchPattern request itself - what are we looking for
   * @param firstOnly return only one (first) satisfying reply
   * @param shareGroup name of share group that we should search through (empty string means "all")
   * @param timeout maximal time in milliseconds to wait for the reply(ies)
   * @return list of replies (SearchReplyItem); null when no replies
   */
  public List search(SearchPattern searchPattern, boolean firstOnly, String shareGroup, int timeout) throws CoreException
  {
    Reporter.startInfo("Searching for '" + searchPattern.getStr() + "'");
    
    SearchRequest request = new SearchRequest();
    request.setSearchPattern(searchPattern);
    request.setFirstOnly(firstOnly);
    request.setShareGroup(shareGroup);
    request.setTimeout(timeout);

    try
    {
      search.request(request);
    }
    catch (SearchException e)
    {
      throw new CoreException(e.getMessage(), e);
    }

    List replies = request.getReplies();
    if (request.getErrCode() == 0 && replies != null && replies.size() > 0)
    {
      SearchReplyItem replyItem = (SearchReplyItem)replies.get(0);
      Reporter.stopInfo("Bundles for '" + searchPattern.getStr() + "' found (e.g. '" + replyItem.getBundleName() + "' at '" + replyItem.getNodeName() + "')");
      return replies;
    }
    else
    {
      Reporter.stopInfo("No bundles for '" + searchPattern.getStr() + "' + found");
      return null;
    }
  }
}
