/*
 * BundleContent.java
 *
 * Created on 23. listopad 2003, 16:18
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import SOFA.SOFAnode.TR.Impl.ComponentInfoImpl;
import SOFA.SOFAnode.TR.ComponentInfo;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the bundle content.
 * One item of persistent storage.
 * <p>
 * The bundle contant contains:
 * <ul>
 * <li>list of components contained in the bundle
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class BundleContent extends XMLStorageItem implements StorageItem, Serializable
{
  private List components;
  
  /** Creates a new instance of BundleContent */
  public BundleContent(String name, File file)
  {
    super(name, file, "bundle_content");
    components = Collections.synchronizedList(new LinkedList());
  }

  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    components.clear();
  }
  
  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;bundle_content&gt;
   *   BundleContent XML format
   * &lt;/bundle_content&gt;
   * </pre>
   * <p>
   * XML format:
   * <p>
   * <pre>
   *  *&lt;component name="string" implementation_version="string"/&gt;
   * </pre>
   */
  public void loadFromXML(Element element) throws SAXException
  {
    synchronized (components)
    {
      components.clear();
      NodeList nl = element.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++)
      {
        Node node = nl.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE && node.getNodeName().compareTo("component") == 0)
        {
          Element e = (Element)node;
          String name = e.getAttribute("name");
          String version = e.getAttribute("implementation_version");
          if (name.length() == 0) throw new SAXException("Missing 'name' attribute in 'component' tag");
          if (version.length() == 0) throw new SAXException("Missing 'implementation_version' attribute in 'component' tag");
          components.add(new ComponentInfoImpl(name, version));
        }
      }
    }
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    synchronized (components)
    {
      Document doc = element.getOwnerDocument();
      Iterator it = components.iterator();
      while (it.hasNext())
      {
        ComponentInfo ci = (ComponentInfo)it.next();
        Element el = doc.createElement("component");
        el.setAttribute("name", ci.getName());
        el.setAttribute("implementation_version", ci.getImplementationVersion());
        element.appendChild(el);
      }
    }      
  }
  
  
  public List getComponents()
  {
    return components;
  }

}
