/*
 * BundleContents.java
 *
 * Created on 23. listopad 2003, 16:16
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import SOFA.SOFAnode.TR.ComponentInfo;

/**
 * Persistent storage of bundle contents (lists of components).
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class BundleContents extends StorageBase
{
  private CompBundleMap compBundleMap;
  
  /** Creates a new instance of BundleContents */
  public BundleContents(File baseDir)
  {
    super(baseDir, "content", true);  //in memory only at the startup, than removed!
    init();
    createCompBundleMap();
  }
  
  protected StorageItem newItem(String name, File file)
  {
    String bundleName = BundleInfo.fileNameToBundleName(name);
    if (bundleName == null) return null;
    else return new BundleContent(bundleName, file);
  }
  
  /**
   * Created component-bundle map that is used to quick look up of the bundle that contains concrete component
   */
  public void createCompBundleMap()
  {
    compBundleMap = new CompBundleMap();
    Iterator it = map.values().iterator();
    while (it.hasNext())
    {
      BundleContent bundleContent = (BundleContent)it.next();
      compBundleMap.addBundleContent(bundleContent);
      bundleContent.setLoaded(false);  //remove from memory
    }
  }
  
  /**
   * Returns component-bundle map
   *
   * @return component-bundle map
   */
  public CompBundleMap getCompBundleMap()
  {
    return compBundleMap;
  }
  
  /**
   * Returns content of the specified bundle
   * The function assures that the returned BundleContent is loaded in the memory.
   *
   * @param bundleName name of bundle
   * @return BundleContent object for the specified bundle; null if not present
   */
  public BundleContent getBundleContent(String bundleName)
  {
    BundleContent bundleContent = (BundleContent)map.get(bundleName);
    if (bundleContent != null && !bundleContent.isLoaded()) bundleContent.loadFromStorage();
    return bundleContent;
  }
  
  /**
   * Deletes bundle content of the specified bundle.
   *
   * @param bundleName name of bundle
   * @return true, if bundle content was found (and deleted)
   */
  public boolean deleteBundleContent(String bundleName)
  {
    synchronized (map)
    {
      BundleContent bundleContent = (BundleContent)map.remove(bundleName);
      if (bundleContent != null)
      {
        bundleContent.deleteFromStorage();
        compBundleMap.deleteBundle(bundleName);
        return true;
      }
      else return false;
    }
  }
  
  /**
   * Adds bundle content for the specified bundle.
   *
   * @param bundleName name of bundle
   * @param components array of the components that the bundle consist of
   * @return bundle content, if bundle content was added (bundleName was correct and there was not already the one with the same name), null otherwise
   */
  public BundleContent addBundleContent(String bundleName, ComponentInfo[] components)
  {
    synchronized (map)
    {
      if (map.get(bundleName) != null) return null;

      String fileName = BundleInfo.bundleNameToFileName(bundleName);
      if (fileName == null) return null;
      BundleContent bundleContent = new BundleContent(bundleName, new File(baseDir, fileName + ".content"));
      List compList = bundleContent.getComponents();
      for (int i = 0; i < components.length; i++)
      {
        compList.add(components[i]);
      }

      compBundleMap.addBundleContent(bundleContent);
      bundleContent.saveToStorage();
      map.put(bundleName, bundleContent);
      return bundleContent;
    }
  }
}
