/*
 * BundleOffers.java
 *
 * Created on 19. bezen 2004, 20:53
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import SOFA.SOFAnode.TR.Impl.ComponentInfoImpl;

/**
 * Persistent storage of bundle offers.
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class BundleOffers extends StorageBase
{
  private CompBundleMap compBundleMap;
  
  /** Creates a new instance of BundleOffers */
  public BundleOffers(File baseDir)
  {
    super(baseDir, "offer", true);  //in memory only at the startup, than removed!
    init();
    createCompBundleMap();
  }
  
  protected StorageItem newItem(String name, File file)
  {
    String bundleName = BundleInfo.fileNameToBundleName(name);
    if (bundleName == null) return null;
    else return new BundleOffer(bundleName, file);
  }
  
  /**
   * Created component-bundle map that is used to quick look up of the bundle offer that contains concrete component
   */
  public void createCompBundleMap()
  {
    compBundleMap = new CompBundleMap();
    Iterator it = map.values().iterator();
    while (it.hasNext())
    {
      BundleOffer bundleOffer = (BundleOffer)it.next();
      compBundleMap.addBundleOffer(bundleOffer);
      bundleOffer.setLoaded(false);  //remove from memory
    }
  }
  
  /**
   * Returns component-bundle map
   *
   * @return component-bundle map
   */
  public CompBundleMap getCompBundleMap()
  {
    return compBundleMap;
  }
  
  /**
   * Returns specified bundle offer
   * The function assures that the returned BundleOffer is loaded in the memory.
   *
   * @param bundleName name of bundle offer
   * @return BundleOffer object for the specified bundle; null if not present
   */
  public BundleOffer getBundleOffer(String bundleName)
  {
    BundleOffer bundleOffer = (BundleOffer)map.get(bundleName);
    if (bundleOffer != null && !bundleOffer.isLoaded()) bundleOffer.loadFromStorage();
    return bundleOffer;
  }
  
  /**
   * Deletes specified bundle offer.
   *
   * @param bundleName name of bundle offer
   * @return true, if bundle offer was found (and deleted)
   */
  public boolean deleteBundleOffer(String bundleName)
  {
    synchronized (map)
    {
      BundleOffer bundleOffer = (BundleOffer)map.remove(bundleName);
      if (bundleOffer != null)
      {
        bundleOffer.deleteFromStorage();
        compBundleMap.deleteBundle(bundleName);
        return true;
      }
      else return false;
    }
  }
  
  /**
   * Adds bundle offer.
   *
   * @param bundleName name of bundle offer
   * @param nodeName name of node that can provide real bundle (that is offering the bundle)
   * @param components array of the components that the bundle consist of
   * @return bundle offer, if bundle offer was added (bundleName was correct and there was not already the one with the same name), null otherwise
   */
  public BundleOffer addBundleOffer(String bundleName, String nodeName, ComponentInfoImpl[] components)
  {
    synchronized (map)
    {
      if (map.get(bundleName) != null) return null;

      String fileName = BundleInfo.bundleNameToFileName(bundleName);
      if (fileName == null) return null;
      BundleOffer bundleOffer = new BundleOffer(bundleName, new File(baseDir, fileName + ".offer"));
      bundleOffer.setNodeName(nodeName);
      List compList = bundleOffer.getComponents();
      for (int i = 0; i < components.length; i++)
      {
        compList.add(components[i]);
      }

      compBundleMap.addBundleOffer(bundleOffer);
      bundleOffer.saveToStorage();
      map.put(bundleName, bundleOffer);
      return bundleOffer;
    }
  }
}
