/*
 * ContractRule.java
 *
 * Created on 3. prosinec 2003, 18:59
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import java.text.*;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the one contract rule
 * (Contract consists of ContractRules - one bussiness contract can cover more
 *  software - ContractRules).
 * <p>
 * ContractRule consists of:
 * <ul>
 * <li>name of the rule
 * <li>test description
 * <li>time validity
 * <li>software - filter of bundles - whis bundles are covered by this ContractRule
 * <li>"push part" - whether to push bundles to the nodes and under which conditions
 *   <ul>
 *   <li>filter of "pushing" nodes (contract is taken only if the bundle arrives from such node)
 *   <li>names of destination nodes and number of licence copies to send
 *   <li>counter of pushes - how many times do the push
 *   </ul>
 * <li>"pull part"
 *   <ul>
 *   <li>filter of nodes that can pull bundles from the node
 *   <li>counter of pulls - how many times can be pulled from this node
 *   </ul>
 * <li>licence
 * </ul>
 * <p>
 * Contract contains several default settings for all ContractRules (if these thing are not filled in ContractRule).
 *  
 * @author  Ladislav Sobr
 */
public class ContractRule implements Cloneable, Serializable
{
  private Contract parentContract;
  
  private String name;
  private String description;
  private Date validFrom;         //can be null
  private Date validTo;           //can be null
  private boolean autoDelete;
  private BundleNameFilter software;
  private boolean isPush;
  private int pushCounter;        //-1 == no counter
  private NodeNameLicList pushDestinationNodes;
  private NodeNameFilter pushNodeFilter;
  private boolean isPull;
  private NodeNameFilter pullNodeFilter;
  private int pullCounter;        //-1 == no counter
  private Licence licence;
  
  /** Creates a new instance of ContractRule */
  public ContractRule(String name, Contract parentContract)
  {
    this.name = name;
    this.parentContract = parentContract;
    reset();
  }

  public Object clone()
  {
    ContractRule clone = null;
    try
    { 
      clone = (ContractRule)super.clone();
    }
    catch (CloneNotSupportedException e)
    { 
      throw new InternalError();
    }
    if (software != null) clone.software = (BundleNameFilter)software.clone();
    if (pushDestinationNodes != null) clone.pushDestinationNodes = (NodeNameLicList)pushDestinationNodes.clone();
    if (pushNodeFilter != null) clone.pushNodeFilter = (NodeNameFilter)pushNodeFilter.clone();
    if (pullNodeFilter != null) clone.pullNodeFilter = (NodeNameFilter)pullNodeFilter.clone();
    return clone;
  }
  
  /**
   * Resets content of the ContractRule.
   */
  protected void reset()
  {
    description = "";
    validFrom = null;
    validTo = null;
    autoDelete = false;
    software = new BundleNameFilter();
    isPush = false;
    pushCounter = -1;
    pushDestinationNodes = null;
    pushNodeFilter = null;
    isPull = false;
    pullNodeFilter = null;
    pullCounter = -1;
    licence = new Licence();
  }
  
  /**
   * Loads the ContractRule from the XML DOM tree.
   * <p>
   * XML format:
   * <p>
   * <pre>
   *  ?&lt;descripton&gt;string&lt;/description&gt;
   *  ?&lt;validity ?from="date" ?to="date" ?autodelete="boolean"/&gt;
   *
   *   &lt;software&gt;
   *     BundleNameFilter XML format
   *   &lt;/software&gt;
   *
   *  ?&lt;push&gt;
   *    ?&lt;push_node_filter&gt;            (presence of this tag overrides "defaults" in contract)
   *       NodeNameFilter XML format
   *     &lt;/push_node_filter&gt;
   *    ?&lt;push_destination_nodes&gt;            (presence of this tag overrides "defaults" in contract)
   *       NodeNameLicList XML format        (presence of "number_of_licences" overrides "number_of_copies" in licence)
   *     &lt;/push_destination_nodes&gt;
   *    ?&lt;counter&gt;integer&lt;/counter&gt;
   *   &lt;/push&gt;
   *
   *  ?&lt;pull&gt;
   *    ?&lt;pull_node_filter&gt;                  (presence of this tag overrides "defaults" in contract)
   *       NodeNameFilter XML format
   *     &lt;/pull_node_filter&gt;
   *    ?&lt;counter&gt;integer&lt;/counter&gt;
   *   &lt;/pull&gt;
   *
   *   &lt;licence&gt;
   *     Licence XML format
   *   &lt;/licence&gt;
   * </pre>
   */
  public void loadFromXML(Element element) throws SAXException
  {
    reset();
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        Element el = (Element)node;
        String nodeName = node.getNodeName();
        
        if (nodeName.compareTo("description") == 0)
        {
          description = XML.getTextContent(el);
        }
        else
        if (nodeName.compareTo("validity") == 0)
        {
          String from = el.getAttribute("from");
          String to = el.getAttribute("to");
          String ad = el.getAttribute("autodelete");
          DateFormat dateFormat = DateFormat.getInstance();
          if (from.length() > 0)
          {
            try
            {
              validFrom = dateFormat.parse(from);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'from' attribute of 'validity' tag", e);
            }
          }
          
          if (to.length() > 0)
          {
            try
            {
              validTo = dateFormat.parse(to);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'to' attribute of 'validity' tag", e);
            }
          }

          if (ad.length() > 0)
          {
            ad = ad.trim().toLowerCase();
            if (ad.compareTo("1") == 0 || ad.compareTo("true") == 0) autoDelete = true;
          }
        }
        else
        if (nodeName.compareTo("software") == 0)
        {
          software.loadFromXML(el);
        }
        else
        if (nodeName.compareTo("push") == 0)
        {
          isPush = true;
          
          NodeList nl2 = el.getChildNodes();
          for (int i2 = 0; i2 < nl2.getLength(); i2++)
          {
            Node node2 = nl2.item(i2);
            if (node2.getNodeType() == Node.ELEMENT_NODE)
            {
              Element el2 = (Element)node2;
              String nodeName2 = node2.getNodeName();

              if (nodeName2.compareTo("push_node_filter") == 0)
              {
                pushNodeFilter = new NodeNameFilter();
                pushNodeFilter.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("push_destination_nodes") == 0)
              {
                pushDestinationNodes = new NodeNameLicList();
                pushDestinationNodes.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("counter") == 0)
              {
                try
                {
                  int l = Integer.parseInt(XML.getTextContent(el2));
                  if (l < 0) l = -1;
                  pushCounter = l;
                }
                catch (NumberFormatException e)
                {
                  throw new SAXException("Invalid 'integer' format of content of 'counter' tag", e);
                }
              }
            }
          }
        }
        else
        if (nodeName.compareTo("pull") == 0)
        {
          isPull = true;
          
          NodeList nl2 = el.getChildNodes();
          for (int i2 = 0; i2 < nl2.getLength(); i2++)
          {
            Node node2 = nl2.item(i2);
            if (node2.getNodeType() == Node.ELEMENT_NODE)
            {
              Element el2 = (Element)node2;
              String nodeName2 = node2.getNodeName();

              if (nodeName2.compareTo("pull_node_filter") == 0)
              {
                pullNodeFilter = new NodeNameFilter();
                pullNodeFilter.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("counter") == 0)
              {
                try
                {
                  int l = Integer.parseInt(XML.getTextContent(el2));
                  if (l < 0) l = -1;
                  pullCounter = l;
                }
                catch (NumberFormatException e)
                {
                  throw new SAXException("Invalid integer format of content of 'counter' tag", e);
                }
              }
            }
          }
        }
        else
        if (nodeName.compareTo("licence") == 0)
        {
          licence.loadFromXML(el);
        }
      }
    }
  }
  
  /**
   * Saves the ContractRule to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Document doc = element.getOwnerDocument();
    Element el;
  
    if (description.length() != 0) XML.newTextElement(element, "description", description);
    
    if (validFrom != null || validTo != null || autoDelete)
    {
      DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM);
      el = doc.createElement("validity");
      if (validFrom != null) el.setAttribute("from", dateFormat.format(validFrom));
      if (validTo != null) el.setAttribute("to", dateFormat.format(validTo));
      if (autoDelete) el.setAttribute("autodelete", "true");
      element.appendChild(el);
    }

    el = doc.createElement("software");
    software.saveToXML(el);
    element.appendChild(el);
    
    if (isPush)
    {
      el = doc.createElement("push");

      if (pushNodeFilter != null)
      {
        Element el2 = doc.createElement("push_node_filter");
        pushNodeFilter.saveToXML(el2);
        el.appendChild(el2);
      }
      
      if (pushDestinationNodes != null)
      {
        Element el2 = doc.createElement("push_destination_nodes");
        pushDestinationNodes.saveToXML(el2);
        el.appendChild(el2);
      }

      if (pushCounter >= 0)
      {
        XML.newTextElement(el, "counter", Integer.toString(pushCounter));
      }

      element.appendChild(el);
    }

    if (isPull)
    {
      el = doc.createElement("pull");

      if (pullNodeFilter != null)
      {
        Element el2 = doc.createElement("pull_node_filter");
        pullNodeFilter.saveToXML(el2);
        el.appendChild(el2);
      }

      if (pullCounter >= 0)
      {
        XML.newTextElement(el, "counter", Integer.toString(pullCounter));
      }

      element.appendChild(el);
    }

    el = doc.createElement("licence");
    licence.saveToXML(el);
    element.appendChild(el);
  }

  public Contract getParentContract()
  {
    return parentContract;
  }
  
  public String getName()
  {
    return name;
  }

  public String getDescription()
  {
    return description;
  }
  
  public boolean getAutoDelete()
  {
    return autoDelete;
  }
  
  public Date getValidFrom()         //can be null
  {
    return validFrom;
  }
  
  public Date getValidTo()           //can be null
  {
    return validTo;
  }
  
  public boolean isValid()
  {
    Date now = Calendar.getInstance().getTime();
    return (validFrom == null || !validFrom.after(now)) && (validTo == null || !validTo.before(now));
  }

  public boolean toBeDeleted()
  {
    return autoDelete && (
       validTo != null && validTo.before(Calendar.getInstance().getTime()) ||
       isPush && pushCounter == 0 && (!isPull || pullCounter <= 0) ||
       isPull && pullCounter == 0 && (!isPush || pushCounter <= 0));
  }
  
  public BundleNameFilter getSoftware()
  {
    return software;
  }
  
  public boolean isPush()
  {
    return isPush;
  }
  
  public int getPushCounter()
  {
    return pushCounter;
  }
  
  public NodeNameLicList getPushDestinationNodes()
  {
    return pushDestinationNodes;
  }
  
  public NodeNameFilter getPushNodeFilter()
  {
    return pushNodeFilter;
  }
  
  public boolean isPull()
  {
    return isPull;
  }
  
  public NodeNameFilter getPullNodeFilter()
  {
    return pullNodeFilter;
  }
  
  public int getPullCounter()
  {
    return pullCounter;
  }
  
  public Licence getLicence()
  {
    return licence;
  }
  
  public void setParentContract(Contract parentContract)
  {
    this.parentContract = parentContract;
  }
  
  public void setName(String name)
  {
    this.name = name;
  }

  public void setDescription(String description)
  {
    this.description = description;
  }
  
  public void setAutoDelete(boolean autoDelete)
  {
    this.autoDelete = autoDelete;
  }
  
  public void setValidFrom(Date validFrom)         //can be null
  {
    this.validFrom = validFrom;
  }
  
  public void setValidTo(Date validTo)           //can be null
  {
    this.validTo = validTo;
  }
  
  public void setSoftware(BundleNameFilter software)
  {
    this.software = software;
  }
  
  public void setPush(boolean isPush)
  {
    this.isPush = isPush;
  }
  
  public void setPushCounter(int pushCounter)
  {
    this.pushCounter = pushCounter;
  }
  
  public void decrementPushCounter()
  {
    if (pushCounter > 0) pushCounter--;
  }
  
  public void setPushDestinationNodes(NodeNameLicList pushDestinationNodes)
  {
    this.pushDestinationNodes = pushDestinationNodes;
  }
  
  public void setPushNodeFilter(NodeNameFilter pushNodeFilter)
  {
    this.pushNodeFilter = pushNodeFilter;
  }
  
  public void setPull(boolean isPull)
  {
    this.isPull = isPull;
  }
  
  public void setPullNodeFilter(NodeNameFilter pullNodeFilter)
  {
    this.pullNodeFilter = pullNodeFilter;
  }

  public void setPullCounter(int pullCounter)
  {
    this.pullCounter = pullCounter;
  }
  
  public void setLicence(Licence licence)
  {
    this.licence = licence;
  }
  
  
  /**
   * @return true if rule can accept "pull"
   */
  public boolean passPull(BundleInfo bundleInfo, NodeInfo nodeInfo, boolean areDefaultsPresent, boolean nodePassedInDefaults)
  {
    if (pullNodeFilter == null && areDefaultsPresent && !nodePassedInDefaults) return false;
    if (!isValid()) return false;
    if (!isPull) return false;
    if (!software.pass(bundleInfo)) return false;
    if (pullNodeFilter != null && !pullNodeFilter.pass(nodeInfo)) return false;
    if (pullCounter == 0) return false;
    return true;
  }
  
  /**
   * Performs "pull" on the rule. Returned Licence can be null,
   * but only if "pull" cannot be made (passPull would fail).
   * <p>
   * The contract is saved to the storage after the rule is modified.
   *
   * @return Licence extracted from contract (contract rule)
   */
  public Licence performPull()
  {
    if (pullCounter == 0) return null;
    if (pullCounter > 0) pullCounter--;
    Licence lic = (Licence)licence.clone();
    if (lic.getText().length() == 0) lic.setText(parentContract.getDefaultLicenceText());
    parentContract.saveToStorage();
    return lic;
  }
  
  /**
   * Can ContractRule accept push - does ConractRule allow it?
   *
   * @param bundleInfo identification of bundle
   * @param nodeInfo identification of destination node
   * @param nodePassedInDefaults was destnination node accepted by "push defaults" in contract?
   * @return true if rule can accept "push"
   */
  public boolean passPush(BundleInfo bundleInfo, NodeInfo nodeInfo, boolean nodePassedInDefaults)
  {
    if (pushDestinationNodes == null && !nodePassedInDefaults) return false;
    if (!isValid()) return false;
    if (!isPush) return false;
    if (!software.pass(bundleInfo)) return false;
    if (pushDestinationNodes != null && pushDestinationNodes.pass(nodeInfo) == null) return false;
    if (pushCounter == 0) return false;
    return true;
  }

  /**
   * Performs "push" on the rule. Returned Licence can be null,
   * but only if "push" cannot be made (passPush would fail).
   * <p>
   * The contract is saved to the storage after the rule is modified.
   *
   * @param nodeInfo identification of destination node
   * @return Licence extracted from contract (contract rule)
   */
  public Licence performPush(NodeInfo nodeInfo)
  {
    if (pushCounter == 0) return null;
    NodeNameLicList.NodeLicPair nlp = null;
    if (pushDestinationNodes != null) nlp = pushDestinationNodes.pass(nodeInfo);
    else nlp = parentContract.getDefaultPushDestinationNodes().pass(nodeInfo);
    if (nlp == null) return null;
    
    Licence lic = (Licence)licence.clone();
    if (nlp.numberOfLicences != Licence.ONE_IMPLICIT_LICENCE) lic.setNumberOfCopies(nlp.numberOfLicences);
    if (lic.getText().length() == 0) lic.setText(parentContract.getDefaultLicenceText());
    if (pushCounter > 0)
    {
      pushCounter--;
      parentContract.saveToStorage();
    }
    return lic;
  }
}
