/*
 * Contracts.java
 *
 * Created on 14. prosinec 2003, 21:32
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import SOFA.SOFAnet.Core.Reporter;

/**
 * Persistent storage of contracts.
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class Contracts extends StorageBase
{
  
  /** Creates a new instance of Contracts */
  public Contracts(File baseDir)
  {
    super(baseDir, "contract", false);
    init();
  }
  
  protected StorageItem newItem(String name, File file)
  {
    return new Contract(name, file);
  }
  
  /**
   * Returns specified contract.
   * The function assures that the returned Contract is loaded in the memory.
   *
   * @param name name of contract
   * @return requested contract; null if not present
   */
  public Contract getContract(String name)
  {
    Contract contract = (Contract)map.get(name);
    if (contract != null && !contract.isLoaded()) contract.loadFromStorage();
    return contract;
  }

  /**
   * Returns contract rule of the contract.
   * The function assures that parent Contract of the returned ContractRule is loaded in the memory.
   *
   * @param contractID contractID od the contract
   * @param contractRule name of rule in the contract
   * @return requested contract rule; null if not present
   */
  public ContractRule getContractRule(String contractID, String contractRule)
  {
    Contract contract = (Contract)map.get(contractID);
    if (contract == null) return null;
    if (!contract.isLoaded()) contract.loadFromStorage();
    ContractRule rule = (ContractRule)contract.getRules().get(contractRule);
    return rule;
  }
  
  /**
   * Deletes specified contract
   *
   * @param name name of contract
   * @return true, if contract was found (and deleted)
   */
  public boolean deleteContract(String name)
  {
    synchronized (map)
    {
      Contract contract = (Contract)map.remove(name);
      if (contract != null)
      {
        contract.deleteFromStorage();
        return true;
      }
      else return false;
    }
  }

  /**
   * Adds contract.
   * <p>
   * Name of contract is identical to Contract ID.
   * <p>
   * <ul>
   * <li>If customizableName == false:
   *   <ul>
   *   <li>name is full name(id) of contract and is not changed in any way
   *   <li>function can fail
   *   </ul>
   * <li>If customizableName == true:
   *   <ul>
   *   <li>name(id) is prefix of the name, postfix is generated as sequent numbers
   *   <li>if name(id) is empty, name will be randomly generated string
   *   <li>function never fail
   *   </ul>
   * </ul>
   * <p>
   * For description of all parameters see Contract class
   *
   * @param name can be null - in that case the name will be generated and function always succeed
   * @param validFrom can be null
   * @param validTo can be null
   * @return contract, if contract was added (there was not already the one with the same name); null otherwise
   */
  public Contract addContract(String name,
                              boolean customizableName,
                              String description,
                              boolean autoDelete,
                              Date validFrom,
                              Date validTo,
                              NodeNameLicList defaultPushDestinationNodes,
                              NodeNameFilter defaultPushNodeFilter,
                              NodeNameFilter defaultPullNodeFilter,
                              String defaultLicenceText,
                              Collection rules)
  {
    synchronized (map)
    {
      if (customizableName || name == null)
      {
        if (name == null || name.length() == 0)
        {
          name = "";
          Random random = new Random();
          for (int i = 0; i < 20; i++) name += Integer.toString(random.nextInt(16), 16);
        }
        
        String postfix = "";
        for (int i = 0; ; i++)
        {
          if (map.get(name + postfix) == null)
          {
            name += postfix;
            break;
          }
          postfix = Integer.toString(i + 1);
        }
      }
      else
      {
        if (map.get(name) != null) return null;
      }

      Contract contract = new Contract(name, new File(baseDir, name + ".contract"));
      contract.setContractID(name);
      contract.setDescription(description);
      contract.setAutoDelete(autoDelete);
      contract.setValidFrom(validFrom);
      contract.setValidTo(validTo);
      contract.setDefaultPushDestinationNodes(defaultPushDestinationNodes);
      contract.setDefaultPushNodeFilter(defaultPushNodeFilter);
      contract.setDefaultPullNodeFilter(defaultPullNodeFilter);
      contract.setDefaultLicenceText(defaultLicenceText);
      
      Iterator it = rules.iterator();
      while (it.hasNext())
      {
        contract.addRule((ContractRule)it.next());
      }
      
      contract.saveToStorage();
      map.put(name, contract);
      return contract;
    }
  }
  
//  
//  
//  /**
//   *  Tries to find a contract and its rule, that can accept "pull" and
//   *  performs "pull" on it.
//   *
//   *  @param bundleInfo describes bundle that should be "pulled"
//   *  @param nodeInfo describes node trying to "pull"
//   *  @return extracted licence or null if no contract(rule) accepts "pull"
//   */
//  public Licence performPull(BundleInfo bundleInfo, NodeInfo nodeInfo)
//  {
//    synchronized (map)
//    {
//      Iterator it = map.values().iterator();
//      while (it.hasNext())
//      {
//        Contract contract = (Contract)it.next();
//        if (!contract.isLoaded()) contract.loadFromStorage();
//        ContractRule contractRule = contract.passPull(bundleInfo, nodeInfo);
//        if (contractRule != null)
//        {
//          //we found the contract rule that can accept "pull"
//          return contractRule.performPull(contract.getDefaultLicenceText());
//        }
//      }
//    }
//    
//    return null;
//  }
  
  /**
   * Performs 'push' action on specified contract and rule.
   *
   * @param bundleName name of bundle
   * @param nodeName exact(!!! as it is stated in contract) name of node, which the bundle should be pushed to
   * @param contractID contractID od the contract
   * @param contractRule name of rule in the contract (must be specified!!!)
   * @return licence for the bundle and node ; null on error
   */
  public Licence performPush(String bundleName, String nodeName, String contractID, String contractRule)
  {
    if (contractID.length() == 0) return null;
    
    synchronized (map)
    {
      Contract contract = (Contract)map.get(contractID);
      if (contract == null) return null;
      if (!contract.isLoaded()) contract.loadFromStorage();
      if (contract.getContractID().compareTo(contractID) != 0)
      {
        Reporter.warning("Incosistent name and contractID of contract. name: '" + contract.getName() + "' contractID: '" + contract.getContractID() + "'");
        return null;
      }
      
      if (!contract.isValid())
      {
        Reporter.warning("Attempt to use Contract that is not valid. contractID: " + contract.getName());
        return null;
      }
      
      ContractRule rule = (ContractRule)contract.getRules().get(contractRule);
      if (rule == null) return null;
      if (!rule.isValid())
      {
        Reporter.warning("Attempt to use ContractRule that is not valid. contractID: " + contract.getName() + " ; contractRule: " + rule.getName());
        return null;
      }
      
      if (!rule.isPush())
      {
        Reporter.warning("Attempt to use ContractRule without 'push' statement. contractID: " + contract.getName() + " ; contractRule: " + rule.getName());
        return null;
      }
      
      BundleNameFilter bundleFilter = rule.getSoftware();
      if (bundleFilter == null || !bundleFilter.pass(bundleName))
      {
        Reporter.warning("Attempt to use ContractRule on impropper bundle. contractID: " + contract.getName() + " ; contractRule: " + rule.getName() + " ; bundleName: " + bundleName);
        return null;
      }

      NodeNameLicList destinations = rule.getPushDestinationNodes();
      if (destinations == null) destinations = contract.getDefaultPushDestinationNodes();
      
      if (destinations == null)
      {
        Reporter.warning("Attempt to use ContractRule for 'push' without destination nodes. contractID: " + contract.getName() + " ; contractRule: " + rule.getName());
        return null;
      }

      NodeNameLicList.NodeLicPair nlp = destinations.find(nodeName);
      if (nlp == null)
      {
        Reporter.warning("Cannot find push destination node in ContractRule. contractID: " + contract.getName() + " ; contractRule: " + rule.getName() + " ; nodeName: " + nodeName);
        return null;
      }

      int pushCounter = rule.getPushCounter();
      if (pushCounter == 0)
      {
        Reporter.warning("Cannot find push destination node in ContractRule. contractID: " + contract.getName() + " ; contractRule: " + rule.getName() + " ; nodeName: " + nodeName);
        return null;
      }
      
      if (pushCounter > 0)
      {
        rule.decrementPushCounter();
        contract.saveToStorage();
      }
      
      Licence licence = (Licence)rule.getLicence().clone();
      
      if (nlp.numberOfLicences != Licence.ONE_IMPLICIT_LICENCE) licence.setNumberOfCopies(nlp.numberOfLicences);
      if (licence.getText().length() == 0) licence.setText(contract.getDefaultLicenceText());
      return licence;
    }      
  }

  /**
   * Tries to find a contract and its rule, that can accept "push".
   *
   * @param bundleInfo describes bundle that should be "pushed"
   * @param nodeInfo destination node
   * @return ContractRule that can accept "push" or null if no such rule exists
   */
  public ContractRule passPush(BundleInfo bundleInfo, NodeInfo nodeInfo)
  {
    synchronized (map)
    {
      Iterator it = map.values().iterator();
      while (it.hasNext())
      {
        Contract contract = (Contract)it.next();
        if (!contract.isLoaded()) contract.loadFromStorage();
        ContractRule contractRule = contract.passPush(bundleInfo, nodeInfo);
        if (contractRule != null) return contractRule;
      }
    }
    
    return null;
  }
  
  /**
   * Tests whether 'pull' action can be done on specified contract.
   *
   * @param bundleInfo "name of bundle"
   * @param nodeInfo node, which requests the bundle
   * @param contractID contractID od the contract
   * @return ContractRule that can accept "pull" or null if no such rule exists
   */
  public ContractRule passPull(BundleInfo bundleInfo, NodeInfo nodeInfo, String contractID)
  {
    if (contractID.length() == 0) return null;
    
    synchronized (map)
    {
      Contract contract = (Contract)map.get(contractID);
      if (contract == null) return null;
      if (!contract.isLoaded()) contract.loadFromStorage();
      if (contract.getContractID().compareTo(contractID) != 0)
      {
        Reporter.warning("Incosistent name and contractID of contract. name: '" + contract.getName() + "' contractID: '" + contract.getContractID() + "'");
        return null;
      }
      return contract.passPull(bundleInfo, nodeInfo);
    }      
  }

  /**
   * Tests whether 'pull' action can be done on specified contract and rule.
   *
   * @param bundleInfo "name of bundle"
   * @param nodeInfo node, which requests the bundle
   * @param contractID contractID od the contract
   * @param contractRule name of rule in the contract
   * @return ContractRule if it can accept "pull" or null if it can't
   */
  public ContractRule passPull(BundleInfo bundleInfo, NodeInfo nodeInfo, String contractID, String contractRule)
  {
    if (contractID.length() == 0) return null;
    
    synchronized (map)
    {
      Contract contract = (Contract)map.get(contractID);
      if (contract == null) return null;
      if (!contract.isLoaded()) contract.loadFromStorage();
      if (contract.getContractID().compareTo(contractID) != 0)
      {
        Reporter.warning("Incosistent name and contractID of contract. name: '" + contract.getName() + "' contractID: '" + contract.getContractID() + "'");
        return null;
      }
      return contract.passPull(bundleInfo, nodeInfo, contractRule);
    }      
  }
  
}
