/*
 * Licence.java
 *
 * Created on 14. prosinec 2003, 17:39
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import java.text.*;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the licence.
 * The class is used either as one item of persistent storage (called licence files)
 * or as the represenation of licence in other data structures.
 * <p>
 * The licence contains:
 * <ul>
 * <li>type of licence: active copies (only presence of "software" in memory is limited) or physical copies (physical presence of the "software" is limited
 * <li>time validity of the licence
 * <li>number of licence copies or flag that this number is not limited (without copies)
 * <li>text of licence
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class Licence extends XMLStorageItem implements StorageItem, Cloneable, Serializable
{
  final static public int TYPE_PHYSICAL = 0;
  final static public int TYPE_ACTIVE = 1;
  final static public int ALL_LICENCES = -2;
  final static public int ONE_IMPLICIT_LICENCE = -1;
  final static public int WITHOUT_COPIES = -3;
  
  private int type;
  private int numberOfCopies;     //WITHOUT_COPIES or integer >= 0
  private Date validFrom;         //can be null
  private Date validTo;           //can be null
  private String text;

  /** Creates a new instance of Licence */
  public Licence(String name, File file)
  {
    super(name, file, "licence");
    reset();
  }
  
  public Licence()
  {
    super("", null, "licence");
    reset();
  }
  
  public Object clone()
  {
    Licence clone = null;
    try
    {
      clone = (Licence)super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      throw new InternalError(); 
    }
    if (validFrom != null) clone.validFrom = (Date)validFrom.clone();
    if (validTo != null) clone.validTo = (Date)validTo.clone();
    return clone;
  }
  
  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    type = TYPE_PHYSICAL;
    numberOfCopies = WITHOUT_COPIES;
    validFrom = null;
    validTo = null;
    text = "";
  }

  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;licence&gt;
   *   Licence XML format
   * &lt;/licence&gt;
   * </pre>
   * <p>
   * XML format:
   * <p>
   * <pre>
   *  ?&lt;validity ?from="date" ?to="date"/&gt;
   *  ?&lt;type&gt;PHYSICAL_COPY or ACTIVE_COPY&lt;/type&gt;
   *  ?&lt;number_of_copies&gt;non-negative-integer&lt;/number_of_copies&gt;
   *  ?&lt;text&gt;string&lt;/text&gt;
   * </pre>
   *
   */
  public void loadFromXML(Element element) throws SAXException
  {
    reset();
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        Element el = (Element)node;
        String nodeName = node.getNodeName();
        
        if (nodeName.compareTo("validity") == 0)
        {
          String from = el.getAttribute("from");
          String to = el.getAttribute("to");
          DateFormat dateFormat = DateFormat.getInstance();
          if (from.length() > 0)
          {
            try
            {
              validFrom = dateFormat.parse(from);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'from' attribute of 'validity' tag", e);
            }
          }
          
          if (to.length() > 0)
          {
            try
            {
              validTo = dateFormat.parse(to);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'to' attribute of 'validity' tag", e);
            }
          }
        }
        else
        if (nodeName.compareTo("type") == 0)
        {
          String t = XML.getTextContent(el).trim().toLowerCase();
          if (t.compareTo("physical_copy") == 0) type = TYPE_PHYSICAL;
          else if (t.compareTo("active_copy") == 0) type = TYPE_ACTIVE;
          else throw new SAXException("Invalid type of licence ('type' tag)");
        }
        else
        if (nodeName.compareTo("number_of_copies") == 0)
        {
          try
          {
            int l = Integer.parseInt(XML.getTextContent(el));
            if (l < 0) throw new NumberFormatException();
            numberOfCopies = l;
          }
          catch (NumberFormatException e)
          {
            throw new SAXException("Invalid 'positive-integer' format of content of 'number_of_copies' tag", e);
          }
        }
        else
        if (nodeName.compareTo("text") == 0)
        {
          text = XML.getTextContent(el);
        }
      }
    }
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Document doc = element.getOwnerDocument();
    Element el;
  
    if (validFrom != null || validTo != null)
    {
      DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM);
      el = doc.createElement("validity");
      if (validFrom != null) el.setAttribute("from", dateFormat.format(validFrom));
      if (validTo != null) el.setAttribute("to", dateFormat.format(validTo));
      element.appendChild(el);
    }
    
    String typeText = "physical_copy";
    switch (type)
    {
      case TYPE_ACTIVE: typeText = "active_copy"; break;
      default:
      case TYPE_PHYSICAL: typeText = "physical_copy"; break;
    }
    
    XML.newTextElement(element, "type", typeText);

    if (numberOfCopies >= 0)
    {
      XML.newTextElement(element, "number_of_copies", Integer.toString(numberOfCopies));
    }
    
    if (text.length() > 0)
    {
      XML.newTextElement(element, "text", text);
    }
  }
    
  public int getType()
  {
    return type;
  }

  public Date getValidFrom()         //can be null
  {
    return validFrom;
  }
  
  public Date getValidTo()           //can be null
  {
    return validTo;
  }
  
  public boolean isValid()
  {
    Date now = Calendar.getInstance().getTime();
    return (validFrom == null || !validFrom.after(now)) && (validTo == null || !validTo.before(now));
  }

  public int getNumberOfCopies()
  {
    return numberOfCopies;
  }
  
  public boolean withCopies()
  {
    return numberOfCopies != WITHOUT_COPIES;
  }

  public String getText()
  {
    return text;
  }

  
  public void setType(int type)
  {
    if (type != TYPE_PHYSICAL && type != TYPE_ACTIVE) type = TYPE_PHYSICAL;
    this.type = type;
  }

  public void setValidFrom(Date validFrom)         //can be null
  {
    this.validFrom = validFrom;
  }
  
  public void setValidTo(Date validTo)           //can be null
  {
    this.validTo = validTo;
  }
  
  public void setNumberOfCopies(int numberOfCopies)
  {
    if (numberOfCopies < 0) numberOfCopies = WITHOUT_COPIES;
    this.numberOfCopies = numberOfCopies;
  }
  
  public void increaseNumberOfCopies(int increment)
  {
    if (numberOfCopies != WITHOUT_COPIES && increment >= 0)
    {
      numberOfCopies += increment;
    }
  }
      
  public void decreaseNumberOfCopies(int decrement)
  {
    if (numberOfCopies != WITHOUT_COPIES && decrement >= 0)
    {
      if (numberOfCopies > decrement) numberOfCopies -= decrement;
      else numberOfCopies = 0;
    }
  }
      
  public void setText(String text)
  {
    this.text = text;
  }

  public boolean isEmpty()
  {
    return type == TYPE_PHYSICAL && numberOfCopies == WITHOUT_COPIES &&
        validFrom == null && validTo == null && text.length() == 0;
  }
}
