/*
 * Licences.java
 *
 * Created on 14. prosinec 2003, 21:28
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;

/**
 * Persistent storage of licences (licence files).
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class Licences extends StorageBase
{
  
  /** Creates a new instance of Licences */
  public Licences(File baseDir)
  {
    super(baseDir, "licence", false);
    init();
  }
  
  protected StorageItem newItem(String name, File file)
  {
    String bundleName = BundleInfo.fileNameToBundleName(name);
    if (bundleName == null) return null;
    else return new Licence(bundleName, file);
  }
  
  /**
   * Returns licence (licence file) for specified bundle
   * The function assures that the returned Licence is loaded in the memory.
   *
   * @param bundleName name of bundle we want the licence file for
   * @return requested licence (licence file); null if not present
   */
  public Licence getLicence(String bundleName)
  {
    Licence licence = (Licence)map.get(bundleName);
    if (licence != null && !licence.isLoaded()) licence.loadFromStorage();
    return licence;
  }

  /**
   * Deletes licence (licence file).
   *
   * @param bundleName name of bundle of which we want to delete licence (licence file)
   * @return true, if licence was found (and deleted)
   */
  public boolean deleteLicence(String bundleName)
  {
    synchronized (map)
    {
      Licence licence = (Licence)map.remove(bundleName);
      if (licence != null)
      {
        licence.deleteFromStorage();
        return true;
      }
      else return false;
    }
  }
  
  /**
   * Adds licence (licence file) for specified bundle.
   *
   * For description of all parameters see Licence class
   *
   * @param bundleName name of bundle
   * @param type see 
   * @param numberOfCopies either Licence.WITHOUT_COPIES or non-negative integer
   * @param validFrom can be null
   * @param validTo can be null
   * @return licence, if licence was added (bundleName was correct and there was not already the one with the same name), null otherwise
   */
  public Licence addLicence(String bundleName,
                            int type,
                            int numberOfCopies,
                            Date validFrom,
                            Date validTo,
                            String text)
  {
    synchronized (map)
    {
      if (map.get(bundleName) != null) return null;
      String fileName = BundleInfo.bundleNameToFileName(bundleName);
      if (fileName == null) return null;
      Licence lic = new Licence(bundleName, new File(baseDir, fileName + ".licence"));
      lic.setType(type);
      lic.setNumberOfCopies(numberOfCopies);
      lic.setValidFrom(validFrom);
      lic.setValidTo(validTo);
      lic.setText(text);

      lic.saveToStorage();
      map.put(bundleName, lic);
      return lic;
    }
  }
}
