/*
 * LocalInfos.java
 *
 * Created on 19. bezen 2004, 0:00
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;

/**
 * Persistent storage of local infos.
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class LocalInfos extends StorageBase
{
  
  /** Creates a new instance of LocalInfos */
  public LocalInfos(File baseDir)
  {
    super(baseDir, "local", true);
    init();
    fixCrashInconsistency();
  }
  
  /**
   * The function corrects local info that can be corrupted by crash of the system.
   * Because one licence copy is taken from "local licence" when the bundle is loaded in the memory,
   * this copy must be returned to local licence, if the "in memory" flag is found during inicialization
   * of this storage.
   */
  protected void fixCrashInconsistency()
  {
    synchronized (map)
    {
      Iterator it = map.values().iterator();
      while (it.hasNext())
      {
        LocalInfo localInfo = (LocalInfo)it.next();
        Licence licence = localInfo.getLicence();
        if (localInfo.isInMemory() && licence.withCopies())
        {
          localInfo.setInMemoryFlag(false);
          licence.increaseNumberOfCopies(1);
          localInfo.saveToStorage();
        }
      }
    }
  }
  
  protected StorageItem newItem(String name, File file)
  {
    String bundleName = BundleInfo.fileNameToBundleName(name);
    if (bundleName == null) return null;
    return new LocalInfo(bundleName, file);
  }
  
  /**
   * Returns local info of specified bundle.
   * The function assures that the returned LocalInfo is loaded in the memory.
   *
   * @param bundleName name of bundle we want the local info for
   * @return requested local info; null if not present
   */
  public LocalInfo getLocalInfo(String bundleName)
  {
    return (LocalInfo)map.get(bundleName);
  }

  /**
   * Deletes local info .
   *
   * @param bundleName name of bundle of which we want to delete local info
   * @return true, if local info was found (and deleted)
   */
  public boolean deleteLocalInfo(String bundleName)
  {
    synchronized (map)
    {
      LocalInfo localInfo = (LocalInfo)map.remove(bundleName);
      if (localInfo != null)
      {
        localInfo.deleteFromStorage();
        return true;
      }
      else return false;
    }
  }

  /**
   * Adds local info for specified bundle.
   * <p>
   * For description of all parameters see LocalInfo class
   *
   * @return local info, if local info was added (bundleName was correct and there was not already the one with the same name), null otherwise
   */
  public LocalInfo addLocalInfo(String bundleName,
                                Licence licence,
                                int state,
                                ShareGroups smShareGroups,
                                NodeNameFilter smNodeFilter,
                                boolean smEquality,
                                String scManager)
  {
    synchronized (map)
    {
      if (map.get(bundleName) != null) return null;

      String fileName = BundleInfo.bundleNameToFileName(bundleName);
      if (fileName == null) return null;
      LocalInfo localInfo = new LocalInfo(bundleName, new File(baseDir, fileName + ".local"));
      localInfo.setLicence(licence);
      localInfo.setState(state);
      localInfo.setSMShareGroups(smShareGroups);
      localInfo.setSMNodeFilter(smNodeFilter);
      localInfo.setSMEquality(smEquality);
      if (localInfo.isShareManager()) localInfo.setSMClients(new NodeNameList());
      localInfo.setSCManager(scManager);
      localInfo.saveToStorage();
      map.put(bundleName, localInfo);
      return localInfo;
    }
  }
  
}
