/*
 * OutputTriggers.java
 *
 * Created on 1. prosinec 2003, 18:15
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;

/**
 * Persistent storage of output triggers.
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class OutputTriggers extends StorageBase
{
  
  /** Creates a new instance of OutputTriggers */
  public OutputTriggers(File baseDir)
  {
    super(baseDir, "ot", true);
    init();
  }
  
  protected StorageItem newItem(String name, File file)
  {
    return new OutputTrigger(name, file);
  }
  
  /**
   * Returns specified output trigger
   * The function assures that the returned OutputTigger is loaded in the memory.
   *
   * @param name name of output trigger
   * @return requested output trigger; null if not present
   */
  public OutputTrigger getOutputTrigger(String name)
  {
    OutputTrigger outputTrigger = (OutputTrigger)map.get(name);
    if (outputTrigger != null && !outputTrigger.isLoaded()) outputTrigger.loadFromStorage();
    return outputTrigger;
  }

  /**
   * Deletes output trigger.
   *
   * @param name name of output trigger
   * @return true, if output trigger was found (and deleted)
   */
  public boolean deleteOutputTrigger(String name)
  {
    synchronized (map)
    {
      OutputTrigger outputTrigger = (OutputTrigger)map.remove(name);
      if (outputTrigger != null)
      {
        outputTrigger.deleteFromStorage();
        return true;
      }
      else return false;
    }
  }
  
  /**
   * Adds output trigger.
   * <p>
   * If contract is not present, set contractID and contractRule to null
   * <p>
   * <ul>
   * <li>If customizableName == false:
   *   <ul>
   *   <li>name is full name of output trigger and is not changed in any way
   *   <li>function can fail
   *   </ul>
   * <li>If customizableName == true:
   *   <ul>
   *   <li>name is prefix of the name, postfix is generated as sequent numbers
   *   <li>if name is empty, prefix is "auto" + randomly generated string
   *   <li>function never fail
   *   </ul>
   * </ul>
   * <p>
   * For description of all parameters see OutputTrigger class
   *
   * @param name can be null - in that case the name will be generated and function always succeed
   * @param validFrom can be null
   * @param validTo can be null
   * @return output trigger, if output trigger was added (there was not already the one with the same name); null otherwise
   */
  public OutputTrigger addOutputTrigger(String name,
                                        boolean customizableName,
                                        String description,
                                        boolean autoDelete,
                                        BundleNameFilter bundleFilter,
                                        NodeNameFilter nodeFilter,
                                        Date validFrom,
                                        Date validTo,
                                        String contractID,
                                        String contractRule)
  {
    synchronized (map)
    {
      if (customizableName || name == null)
      {
        if (name == null || name.length() == 0)
        {
          name = "auto";
          Random random = new Random();
          for (int i = 0; i < 10; i++) name += Integer.toString(random.nextInt(16), 16);
        }
        
        String postfix = "";
        for (int i = 0; ; i++)
        {
          if (map.get(name + postfix) == null)
          {
            name += postfix;
            break;
          }
          postfix = Integer.toString(i + 1);
        }
      }
      else
      {
        if (map.get(name) != null) return null;
      }

      OutputTrigger outputTrigger = new OutputTrigger(name, new File(baseDir, name + ".ot"));
      outputTrigger.setDescription(description);
      outputTrigger.setAutoDelete(autoDelete);
      outputTrigger.setBundleFilter(bundleFilter);
      outputTrigger.setNodeFilter(nodeFilter);
      outputTrigger.setValidFrom(validFrom);
      outputTrigger.setValidTo(validTo);
      outputTrigger.setContract(contractID, contractRule);
      
      outputTrigger.saveToStorage();
      map.put(name, outputTrigger);
      return outputTrigger;
    }
  }
}