/*
 * RMISearchClient.java
 *
 * Created on 29. duben 2004, 16:09
 */

package SOFA.SOFAnet.Search.RMI;

import SOFA.SOFAnet.Search.*;
import SOFA.SOFAnet.Repository.NodeInfo;
import java.util.*;

/**
 * RMI implementation of SearchInterface - client interface of search subsytem.
 *
 * @author  Ladislav Sobr
 */
public class RMISearchClient implements SearchInterface
{
  private RMISearchCenter center;
  
  /** Creates a new instance of RMISearchClient */
  public RMISearchClient(RMISearchCenter center)
  {
    this.center = center;
  }
  
  /**
   * Sends reply to the requesting node.
   * <p>
   * If the connection to the node exists it is used. If not, the conenction is created temporarily.
   */
  public void reply(SearchReply reply) throws SearchException
  {
    String requestingNodeName = reply.getRequestID().getSourceNode();
    
    NodeInfo requestingNodeInfo = new NodeInfo();
    try
    {
      requestingNodeInfo.setNodeName(requestingNodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new SearchException("Invalid name of SOFA node: " + requestingNodeName, e);
    }
    
    //try to find existing connection to the requesting node and use it to send reply
    List rmiSearchConnections = center.getRMISearchConnections();
    synchronized (rmiSearchConnections)
    {
      Iterator it = rmiSearchConnections.iterator();
      while (it.hasNext())
      {
        RMISearchConnection connection = (RMISearchConnection)it.next();
        if (connection.isNode(requestingNodeName))
        {
          try
          {
            connection.sendReply(reply);
          }
          catch (RMISearchException e)
          {
            throw new SearchException(e.getMessage(), e);
          }
          
          return;
        }
      }
    }
    
    //we didn't find the existing connection to the requesting node, create the new one only for reply
    RMISearchConnection connection = new RMISearchConnection(requestingNodeInfo, null, true);

    try
    {
      connection.sendReply(reply);
    }
    catch (RMISearchException e)
    {
      throw new SearchException(e.getMessage(), e);
    }
    
  }
  
  /**
   * Sends the search request to the search network.
   * <p>
   * The collector for the request is created. The request is sent by this collector that is also responsible
   * for collecting replies. When the call of collector finishes, the collector is destroyed.
   */
  public void request(SearchRequest request) throws SearchException
  {
    RMISearchCollector collector = center.newCollector();
    
    collector.setup(request.getSearchPattern(), request.getFirstOnly(), request.getShareGroup(), request.getTimeout());
    List replies = collector.request();
    center.removeCollector(collector.getRequestMark());
    
    request.setReplies(replies);
    request.setErrCode(0);
  }
  
}
