/*
 * SearchConfig.java
 *
 * Created on 29. duben 2004, 10:42
 */

package SOFA.SOFAnet.Search.RMI;

import SOFA.SOFAnet.Repository.XMLStorageItem;
import SOFA.SOFAnet.Repository.Repository;
import SOFA.SOFAnet.Repository.NodeNameFilter;
import SOFA.SOFAnet.Repository.ShareGroups;
import SOFA.SOFAnet.Repository.NodeInfo;
import SOFA.SOFAnet.Core.Reporter;
import SOFA.Util.XML;
import java.io.*;
import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Loader of configuration of RMI implementation of search subsystem.
 *
 * @author  Ladislav Sobr
 */
public class RMISearchConfig
{
  private File cfgFile;
  private List rmiSearchConnections;
  private RMISearchConfiguration configuration;
  
  /** Creates a new instance of SearchConfig */
  public RMISearchConfig(Repository rep, List rmiSearchConnections)
  {
    cfgFile = new File(rep.getInfoBaseDir(), "search_configuration.xml");
    this.rmiSearchConnections = rmiSearchConnections;
    configuration = new RMISearchConfiguration();
  }

  /**
   * Loads configuration from the XML file.
   * <p>
   * Search configuration XML format:
   * <p>
   * <pre>
   * &lt;search_configuration&gt;
   *  ?&lt;node_filter&gt;
   *     NodeNameFilter XML format
   *   &lt;/node_filter&gt;
   *   &lt;connections&gt;
   *    *&lt;connection node="string"&gt;
   *      ?&lt;share_groups&gt;
   *         ShareGroups XML format
   *       &lt;/share_groups&gt;
   *     &lt;/connection&gt;
   *   &lt;/connections&gt;
   * &lt;/search_configuration&gt;
   * </pre>
   *
   * @param start true when loading the configuration first time
   */
  public void loadFromStorage(boolean start)
  {
    configuration.reset();
    
    try
    {
      Document doc = XML.parse(cfgFile);
      Element root = doc.getDocumentElement();
      
      NodeList nl = root.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++)
      {
        Node node = nl.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE)
        {
          Element el = (Element)node;
          String nodeName = node.getNodeName();

          if (nodeName.compareTo("node_filter") == 0)
          {
            NodeNameFilter nodeFilter = new NodeNameFilter();
            nodeFilter.loadFromXML(el);
            configuration.setNodeFilter(nodeFilter);
          }
          else
          if (nodeName.compareTo("connections") == 0)
          {
            NodeList nl2 = el.getChildNodes();
            for (int i2 = 0; i2 < nl2.getLength(); i2++)
            {
              Node node2 = nl2.item(i2);
              if (node2.getNodeType() == Node.ELEMENT_NODE)
              {
                Element el2 = (Element)node2;
                String nodeName2 = node2.getNodeName();

                if (nodeName2.compareTo("connection") == 0)
                {
                  ShareGroups shareGroups = null;
                  String searchNodeName = el2.getAttribute("node");

                  if (searchNodeName.length() == 0)
                  {
                    throw new SAXException("Missing node name (attribute 'node' of tag 'connection')");
                  }
                  
                  NodeInfo nodeInfo = new NodeInfo();
                  try
                  {
                    nodeInfo.setNodeName(searchNodeName);
                  }
                  catch (NodeInfo.InvalidNodeNameException e)
                  {
                    Reporter.error("Invalid name of RMI search node: " + nodeName, e);
                    nodeInfo = null;
                  }
                  
                  NodeList nl3 = el2.getChildNodes();
                  for (int i3 = 0; i3 < nl3.getLength(); i3++)
                  {
                    Node node3 = nl3.item(i3);
                    if (node3.getNodeType() == Node.ELEMENT_NODE)
                    {
                      Element el3 = (Element)node3;
                      String nodeName3 = node3.getNodeName();

                      if (nodeName3.compareTo("share_groups") == 0)
                      {
                        shareGroups = new ShareGroups();
                        shareGroups.loadFromXML(el3);
                      }
                    }
                  }
                  
                  configuration.getConnections().add(new RMISearchConfiguration.Connection(searchNodeName, shareGroups != null ? (ShareGroups)shareGroups.clone() : null));
                  if (nodeInfo != null)
                  {
                    boolean connectionFound = false;
                    if (!start)
                    {
                      //do we have this connection?
                      synchronized (rmiSearchConnections)
                      {
                        Iterator it = rmiSearchConnections.iterator();
                        while (it.hasNext())
                        {
                          RMISearchConnection connection = (RMISearchConnection)it.next();
                          if (connection.isNode(nodeName))
                          {
                            connectionFound = true;
                            break;
                          }
                        }
                      }
                    }
                    
                    if (!connectionFound) rmiSearchConnections.add(new RMISearchConnection(nodeInfo, shareGroups, false));
                  }
                }
              }
            }
          }
        }
      }
    }
    catch (IOException e)
    {
      Reporter.error("Error while loading (RMI) search_configuration XML format from file " + cfgFile, e);
    }
    catch (SAXException e)
    {
      Reporter.error("Error while parsing (RMI) search_configuration XML format from file " + cfgFile, e);
    }
  }

  public void saveToStorage()
  {
    try
    {
      Document doc = XML.newDocument();
      Element root = doc.createElement("search_configuration");
      doc.appendChild(root);
      
      Element el;
      NodeNameFilter nodeFilter = configuration.getNodeFilter();
      if (nodeFilter != null)
      {
        el = doc.createElement("node_filter");
        nodeFilter.saveToXML(el);
        root.appendChild(el);
      }
      
      el = doc.createElement("connections");
      
      List connections = configuration.getConnections();
      synchronized (connections)
      {
        Iterator it = connections.iterator();
        while (it.hasNext())
        {
          RMISearchConfiguration.Connection connection = (RMISearchConfiguration.Connection)it.next();
          Element el2 = doc.createElement("connection");
          el2.setAttribute("node", connection.getNodeName());
          ShareGroups shareGroups = connection.getShareGroups();
          if (shareGroups != null)
          {
            Element el3 = doc.createElement("share_groups");
            shareGroups.saveToXML(el3);
            el2.appendChild(el3);
          }
          el.appendChild(el2);
        }
      }
      
      root.appendChild(el);
      
      XML.dump(cfgFile, root);
    }
    catch (IOException e)
    {
      Reporter.error("Saving of (RMI) search_configuration XML format to file '" + cfgFile + "' failed", e);
    }
    
  }
  
  public NodeNameFilter getNodeFilter()
  {
    return configuration.getNodeFilter();
  }
  
  public RMISearchConfiguration getConfiguration()
  {
    return configuration;
  }
  
  public void setConfiguration(RMISearchConfiguration configuration)
  {
    this.configuration = (RMISearchConfiguration)configuration.clone();
    saveToStorage();
  }
  
}
