/*
 * RMISearchConnection.java
 *
 * Created on 29. duben 2004, 10:52
 */

package SOFA.SOFAnet.Search.RMI;

import SOFA.SOFAnet.Search.*;
import SOFA.SOFAnet.Repository.NodeInfo;
import SOFA.SOFAnet.Repository.ShareGroups;
import SOFA.SOFAnet.Core.Reporter;
import java.rmi.RemoteException;
import java.util.*;

/**
 * Search connection to one node (one-way connection).
 * Every connection can by accompanied by list of share groups - if the search request contains search group,
 * this connection will be used only if this group is present among connection's share groups.
 * <p>
 * Search requests and "connect" commands are queued and sent by asynchronously by extra thread,
 * search replies are sent synchronously by calling thread.
 *
 * @author  Ladislav Sobr
 */
public class RMISearchConnection
{
  private NodeInfo nodeInfo;
  private ShareGroups shareGroups;
  private RMISearchServerHolder serverHolder;
  private Thread asyncThread;
  private List asyncRequests;
  private String myNodeName;
  public boolean connected;
  
  final static private int OP_STOP = 0;
  final static private int OP_REQUEST = 1;
  final static private int OP_CONNECT = 2;
  
  private static class AsyncRequest
  {
    public int op;
    public SearchRequestID requestID;
    public SearchPattern searchPattern;
    public boolean firstOnly;
    public String shareGroup;
    
    AsyncRequest(int op, SearchRequestID requestID, SearchPattern searchPattern, boolean firstOnly, String shareGroup)
    {
      this.op = op;
      this.requestID = requestID;
      this.searchPattern = searchPattern;
      this.firstOnly = firstOnly;
      this.shareGroup = shareGroup;
    }
  }
  
  /** Creates a new instance of RMISearchConnection */
  public RMISearchConnection(NodeInfo nodeInfo, ShareGroups shareGroups, boolean forReplyOnly)
  {
    this.nodeInfo = nodeInfo;
    this.shareGroups = shareGroups;
    
    serverHolder = new RMISearchServerHolder();
    serverHolder.setNode(nodeInfo);

    myNodeName = NodeInfo.getLocalNodeName();
    
    asyncRequests = new LinkedList();
    
    if (!forReplyOnly)
    {
      asyncThread = new Thread()
                      {
                        public void run()
                        {
                          asyncThreadProcedure();
                        }
                      };

      asyncThread.start();
    }
    
    connected = false;
  }
  
  private void asyncThreadProcedure()
  {
    for (;;)
    {
      AsyncRequest request;
      
      synchronized (asyncRequests)
      {
        while (asyncRequests.isEmpty())
        {
          try
          {
            asyncRequests.wait();
          }
          catch (InterruptedException e)
          {
          }
        }
        
        request = (AsyncRequest)asyncRequests.remove(0);
      }

      //perform request
      
      switch (request.op)
      {
      case OP_STOP:
        return;
        
      case OP_CONNECT:
        {
          RMISearchInterface server = null;
          try
          {
            server = serverHolder.getServer(true);
          }
          catch (RMISearchException e)
          {
            Reporter.error("Connection to RMI search server at node '" + nodeInfo.getName() + "' failed", e);
          }

          if (server != null)
          {
            try
            {
              server.connect(myNodeName, shareGroups);
            }
            catch (RemoteException e)
            {
              serverHolder.releaseServer(true);
              Reporter.error("Connection to RMI search server at node '" + nodeInfo.getName() + "' failed", e);
            }
            catch (RMISearchException e)
            {
              serverHolder.releaseServer(true);
              Reporter.error("Connection to RMI search server at node '" + nodeInfo.getName() + "' failed", e);
            }
          }
        }
        break;
        
      case OP_REQUEST:
        if (shareGroups == null || request.shareGroup.length() == 0 || shareGroups.contains(request.shareGroup))
        {
          boolean doConnect = !serverHolder.isServer();
          
          RMISearchInterface server = null;
          try
          {
            server = serverHolder.getServer(true);  //can return null if there was an error recently (e.g. node is not accessible)
          }
          catch (RMISearchException e)
          {
            Reporter.error("Connection to RMI search server at node '" + nodeInfo.getName() + "' failed", e);
          }
          
          if (server != null)
          {
            if (doConnect)
            {
              try
              {
                server.connect(myNodeName, shareGroups);
              }
              catch (RemoteException e)
              {
                server = null;
                serverHolder.releaseServer(true);
                Reporter.error("Connection to RMI search server at node '" + nodeInfo.getName() + "' failed", e);
              }
              catch (RMISearchException e)
              {
                server = null;
                serverHolder.releaseServer(true);
                Reporter.error("Connection to RMI search server at node '" + nodeInfo.getName() + "' failed", e);
              }
            }
          }
          
          if (server != null)
          {
            try
            {
              server.request(myNodeName, request.requestID, request.searchPattern, request.firstOnly, request.shareGroup);
            }
            catch (RemoteException e)
            {
              serverHolder.releaseServer(true);
              Reporter.error("Sending (RMI) search request to node '" + nodeInfo.getName() + "' failed", e);
            }
            catch (RMISearchException e)
            {
              serverHolder.releaseServer(true);
              Reporter.error("Sending (RMI) search request to node '" + nodeInfo.getName() + "' failed", e);
            }
          }
        }
        break;
      }
    }
  }
  
  public void sendRequest(SearchRequestID requestID, SearchPattern searchPattern, boolean firstOnly, String shareGroup)
  {
    synchronized (asyncRequests)
    {
      asyncRequests.add(new AsyncRequest(OP_REQUEST, requestID, searchPattern, firstOnly, shareGroup));
      asyncRequests.notify();
    }
  }
  
  public void connect()
  {
    synchronized (asyncRequests)
    {
      asyncRequests.add(new AsyncRequest(OP_CONNECT, null, null, false, null));
      asyncRequests.notify();
    }
    
    connected = true;
  }
  
  public void stop()
  {
    synchronized (asyncRequests)
    {
      asyncRequests.add(new AsyncRequest(OP_STOP, null, null, false, null));
      asyncRequests.notify();
    }
  }
  
  public boolean isNode(String nodeName)
  {
    return nodeInfo.getName().compareTo(nodeName) == 0;
  }
  
  public boolean isConnected()
  {
    return connected;
  }
  
  public void sendReply(SearchReply searchReply) throws RMISearchException
  {
    RMISearchInterface server = serverHolder.getServer(false);

    try
    {
      server.reply(myNodeName, searchReply);
    }
    catch (RemoteException e)
    {
      serverHolder.releaseServer(true);
      throw new RMISearchException(e.getMessage(), e);
    }
    catch (RMISearchException e)
    {
      serverHolder.releaseServer(true);
      throw e;
    }
  }
}
