/*
 * RMISearchServerHolder.java
 *
 * Created on 29. duben 2004, 13:26
 */

package SOFA.SOFAnet.Search.RMI;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import SOFA.SOFAnet.Repository.NodeInfo;
import java.util.Date;
import java.util.Calendar;

/**
 * The object of this class is responsible for connecting to remote RMI server and holding its interface.
 * <p>
 * It can be configured not to connect to the remote server for specific time if the previous attempt
 * was not successfull.
 *
 * @author  Ladislav Sobr
 */
public class RMISearchServerHolder
{
  private final long tryPeriod = 30000;       //how often we try to connect to the server (in milliseconds)
  
  private RMISearchInterface server;
  private NodeInfo nodeInfo;
  private Date lastError;
  
  /** Creates a new instance of RMISearchServerHolder */
  public RMISearchServerHolder()
  {
    lastError = new Date(0);
  }
  
  /** Setups the node to connect to */
  public void setNode(NodeInfo nodeInfo)
  {
    this.nodeInfo = nodeInfo;
  }
  
  /**
   * Returns interface to the RMI server on the remote node.
   * @param justTry if true, the function can return null if the previous attempt recently failed
   */
  public synchronized RMISearchInterface getServer(boolean justTry) throws RMISearchException
  {
    if (nodeInfo == null) throw new RMISearchException("Not specified node");
    
    if (server != null) return server;
    
    if (justTry)
    {
      Date time = new Date(System.currentTimeMillis() - tryPeriod);
      if (lastError.after(time)) return null;      //it is to early to look for another lookup
    }

    try
    {
      server = (RMISearchInterface)Naming.lookup("rmi://" + nodeInfo.getAddressAndPort() + "/SOFAnet/RMISearchServer");
    }
    catch (NotBoundException e)
    {
      releaseServer(true);
      if (justTry) return null;
      else throw new RMISearchException(e.getMessage(), e);
    }
    catch (MalformedURLException e)
    {
      releaseServer(true);
      if (justTry) return null;
      else throw new RMISearchException(e.getMessage(), e);
    }
    catch (RemoteException e)
    {
      releaseServer(true);
      if (justTry) return null;
      else throw new RMISearchException(e.getMessage(), e);
    }    
    
    return server;
  }
  
  /**
   * Releases interface of the remote RMI server
   * @param error do we release the server because an error occured ?
   */
  public void releaseServer(boolean error)
  {
    server = null;
    if (error) lastError = new Date(System.currentTimeMillis());
  }
  
  /**
   * @return true if the object holds active server
   */
  public boolean isServer()
  {
    return server != null;
  }
}
