/**
 * DistributeDialog.java is a part of the SOFA project.
 * This file was created by pepan on 18.4.2003.
 */
package SOFA.SOFAnode.InOut.Impl;

import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JScrollPane;

import SOFA.SOFAnode.InOut.InOut2Client;
import SOFA.SOFAnode.TR.ComponentInfo;

/**
 * A dialog showing details of the chosen action (push/pull components) which should be
 * performed on selected SOFA nodes. It also contains a button to confirmate the chosen action.
 * <p>Note for developers: Just before the distribution of components between two SOFA nodes,
 * a connection between their InOut parts is made. Since the client knows connector names to
 * their InOut2Client interfaces only, the first type of connectors must be <i>derived</i> from
 * the second one. To achieve this, just a simple implementation is provided - both names differ in
 * a piece of text only. Instead of the string 'InOut2ClientRMISkel', the connector name contains
 * 'InOut2InOutRMISkel'.
 * @author Petr Panuska
 */
class DistributeDialog extends JDialog implements ActionListener {

  /**
   * A list of components to be distributed.
   */
  private JList components;

  /**
   * A list of target SOFA nodes.
   */
  private JList nodes;

  /**
   * A button to confirm the selected action.
   */
  private JButton yesButton;

  /**
   * A button to refuse the selected action.
   */
  private JButton noButton;

  /**
   * A set of components which are to be distributed. Each key is a {@link SOFANode}
   * where the component can be obtained from, each value is a HashSet of {@link ComponentInfo components}.
   */
  private Map from;

  /**
   * A set of SOFA nodes where the components are to be distributed.
   */
  private Set to;

  /**
   * If <code>true</code>, the push method is used; pull method otherwise.
   */
  private boolean pushMethod;

  /**
   * If <code>true</code>, only the description of chosen components is distributed.
   */
  private boolean descriptionOnly;

  /**
   *
   * @param from every key is a {@link SOFANode}, every value is a {@link HashSet} with
   * {@link ComponentInfo components} to be distributed.
   * @param to set of {@link SOFANode SOFAnodes} where the components are to be distributed to.
   * @param pushMethod <code>true</code> when the push method is to be employed, <code>false</code> otherwise.
   * @param descriptionOnly if <code>true</code>, the description of chosen components is distributed only.
   */
  DistributeDialog (Map from, Set to, boolean pushMethod, boolean descriptionOnly) {
    this.from = from;
    this.to = to;
    this.pushMethod = pushMethod;
    this.descriptionOnly = descriptionOnly;

    Vector v = new Vector();
    for (Iterator iterator = from.keySet().iterator(); iterator.hasNext();) {
      SOFANode node = (SOFANode) iterator.next();
      Set set = (Set) from.get(node);   // set of components to be distributed from the node
//      set = getTransitiveClosure(set);
      for (Iterator iter2 = set.iterator(); iter2.hasNext();) {
        ComponentInfo info = (ComponentInfo) iter2.next();
        v.add(node.getName() + " - " + info.getName() + "[" + info.getImplementationVersion() + "]");
      }
    }
    components = new JList(v);

    nodes = new JList(to.toArray());

    yesButton = new JButton("Yes");
    yesButton.addActionListener(this);

    noButton = new JButton("No");
    noButton.addActionListener(this);

    GridBagLayout gb = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();
    Container panel = getContentPane();
    panel.setLayout(gb);

    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    c.weighty = 1.0;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.insets = new Insets(1, 1, 1, 1);

    JLabel question;
    question = new JLabel("Really distribute " +
      ((descriptionOnly) ? "the description of " : "") +
      "listed components to nodes below? " +
      ((pushMethod) ? "(Push method)" : ("Pull method"))
    );
    gb.setConstraints(question, c);
    panel.add(question);

    JComponent comp = new JScrollPane(components);
    gb.setConstraints(comp, c);
    panel.add(comp);

    comp = new JScrollPane(nodes);
    gb.setConstraints(comp, c);
    panel.add(comp);

    c.fill = GridBagConstraints.NONE;
    c.gridwidth = GridBagConstraints.RELATIVE;
    c.weighty = 0.0;
    c.anchor = GridBagConstraints.EAST;
    gb.setConstraints(yesButton, c);
    panel.add(yesButton);

    c.gridwidth = GridBagConstraints.REMAINDER;
    c.anchor = GridBagConstraints.WEST;
    gb.setConstraints(noButton, c);
    panel.add(noButton);

    setDefaultCloseOperation(DISPOSE_ON_CLOSE);
    setModal(true);
    pack();
    setVisible(true);
  }

  /**
   * Called when a user presses a button. If the user refuses to perform the action,
   * it just disposes the window. If the user confirms the action, the action is
   * performed immediately.
   * @param e the emitted event.
   */
  public void actionPerformed (ActionEvent e) {
    if (e.getSource() == noButton) {
      dispose();
    }
    if (e.getSource() == yesButton) {
      if (pushMethod) {
        String[] nodes = new String[to.size()]; // where the components distribute to
        int i = 0;
        for (Iterator iterator = to.iterator(); iterator.hasNext();) {
          SOFANode node = (SOFANode) iterator.next();
          // todo current implementation counts with name relation in the following way:
          // the name of the InOut2InOut connector reference of a SOFA node differs from
          // the name of the InOut2Client connection reference of the same SOFA node
          // just in the following string:
          nodes[i++] = node.getReference().toString().replaceAll("InOut2ClientRMISkel", "InOut2InOutRMISkel");
        }
        Set fromSet = from.keySet();
        for (Iterator iterator = fromSet.iterator(); iterator.hasNext();) { // for all source nodes
          SOFANode node = (SOFANode) iterator.next();
          InOut2Client destInOut = node.getInOut();
          if (destInOut != null) {
            HashSet comps = (HashSet) from.get(node);
            String[] compsFullNames = new String[comps.size()];
            int i2 = 0;
            for (Iterator iter2 = comps.iterator(); iter2.hasNext();) {       // for all components at a source node
              ComponentInfo info = (ComponentInfo) iter2.next();
              compsFullNames[i2++] = info.getName() + "[" + info.getImplementationVersion() + "]";
            }
            if (descriptionOnly) {
              destInOut.distributeOffers(compsFullNames, nodes);
            } else {
              destInOut.distributeComponents(compsFullNames, nodes);
            }
          }
        }
      } else { // pull method
        for (Iterator iter1 = to.iterator(); iter1.hasNext();) {  // for all dest nodes
          SOFANode destNode = (SOFANode) iter1.next();
          InOut2Client destInOut = destNode.getInOut();
          if (destInOut != null) {
            Set fromSet = from.keySet();
            for (Iterator iter2 = fromSet.iterator(); iter2.hasNext();) { // for all source nodes
              SOFANode sourceNode = (SOFANode) iter2.next();
              HashSet comps = (HashSet) from.get(sourceNode);

              String[] compsFullNames = new String[comps.size()];
              int i2 = 0;
              for (Iterator iter3 = comps.iterator(); iter3.hasNext();) {       // for all components at the source node
                ComponentInfo info = (ComponentInfo) iter3.next();
                compsFullNames[i2++] = info.getName() + "[" + info.getImplementationVersion() + "]";
              }
              // todo current implementation counts with name relation in the following way:
              // the name of the InOut2InOut connector reference of a SOFA node differs from
              // the name of the InOut2Client connection reference of the same SOFA node
              // just in the following string:
              String node = sourceNode.getReference().toString().replaceAll("InOut2ClientRMISkel", "InOut2InOutRMISkel");
              if (descriptionOnly) {
                destInOut.obtainOffers(compsFullNames, node);
              } else {
                destInOut.obtainComponents(compsFullNames, node);
              }
            }
          }
        } // dest nodes
      } // pull method
    } // yes button
    dispose();
  }
}
