/**
 * InOutImpl.java is a part of the SOFA project.
 * This file was created by pepan on 19.3.2003.
 */
package SOFA.SOFAnode.InOut.Impl;

import java.util.HashMap;

import SOFA.Connector.ConnectorException;
import SOFA.Connector.Reference;
import SOFA.SOFAnode.InOut.Bundle;
import SOFA.SOFAnode.InOut.InOut2Client;
import SOFA.SOFAnode.InOut.InOut2InOut;
import SOFA.SOFAnode.InOut.InOutException;
import SOFA.SOFAnode.InOut.Connector.InOut2InOutConnector;
import SOFA.SOFAnode.TR.ComponentInfo;
import SOFA.SOFAnode.TR.TR2InOut;
import SOFA.SOFAnode.TR.Connector.TR2InOutConnector;
import SOFA.SOFAnode.TR.Impl.ComponentInfoImpl;

/**
 * Provides services for storing bundles of components to TR and for retrieving them from TR.
 * @author Petr Panuska
 */
public class InOutImpl implements InOut2InOut, InOut2Client {

  /**
   * The interface between the InOut part of this SOFA node and the Template Repository part of the same node.
   */
  private TR2InOut tr;

  /**
   * A list of connections to other InOuts. Each key is a URL of a SOFA node,
   * each value is the reference to the InOut interface of this SOFA node.
   */
  private HashMap inOutConn;

  /**
   * Creates the InOut server and connects it to TR.
   * @param ref the connector to TR.
   */
  public InOutImpl (Reference ref) {
    try {
      tr = (TR2InOut) TR2InOutConnector.createClt(ref);
    } catch (ConnectorException e) {
      e.printStackTrace();
      System.exit(-1);
    }
    inOutConn = new HashMap();
  }

  /**
   * Pulls a bundle packing the specified components from this SOFA node.
   * @param descs the list of components whose description is going to be pulled.
   * @param comps the list of components which are going to be pulled (description including
   * component files as well)
   * @param inferiors
   * @return a bundle containing required components.
   * @throws InOutException if some error occured (specified components not present, etc.)
   */
  public Bundle pullBundle (ComponentInfo[] descs, ComponentInfo[] comps, boolean inferiors) throws InOutException {
    return tr.getBundle(descs, comps, inferiors);
  }

  /**
   * Pushes the specified bundle (packing components) to this SOFA node.
   * @param bundle the bundle which is being pushed.
   * @throws InOutException
   */
  public void pushBundle (Bundle bundle) throws InOutException {
    tr.storeBundle(bundle);
  }

  /**
   * Gets a connection to InOut server of a given SOFA node. If the connection
   * does not exist, it is created first.
   * @param sofaNode a string representation of a reference to an InOut2InOut connector of a SOFA node.
   * @return an interface to the InOut part of the SOFA node.
   * @see Reference#fromString(String)
   */
  private InOut2InOut getConnection (String sofaNode) {
    InOut2InOut inOut;
    inOut = (InOut2InOut) inOutConn.get(sofaNode);  // get a previously created connection
    if (inOut == null) {  // the connection has not been created yet
      try {
        inOut = (InOut2InOut) InOut2InOutConnector.createClt(Reference.fromString(sofaNode)); // create it
      } catch (ConnectorException e) {
        throw (new InOutException("Could not connect to " + sofaNode + "\n", e));
      }
      inOutConn.put(sofaNode, inOut);  // store the new created connection
    }
    return inOut;
  }

  /**
   * Distributes a bundle to the given SOFA node.
   * @param bundle the bundle which is going to be distributed.
   * @param sofaNode a string representation of a reference to an InOut2InOut connector of a SOFA node.
   * @throws InOutException when an error occurs.
   */
  private void distributeComponent (Bundle bundle, String sofaNode) throws InOutException {
    InOut2InOut inOut = getConnection(sofaNode);
    inOut.pushBundle(bundle);
  }

  /**
   * Translates a given array of component full names (name[version]) into an array of
   * {@link ComponentInfoImpl} objects. It creates new instances of these objects.
   * @param fullNames the array of component full names.
   * @return an array of component information objects.
   */
  private ComponentInfoImpl[] fullNamesToComponentInfos (String[] fullNames) {
    ComponentInfoImpl[] comps = null;
    if (fullNames != null) {
      comps = new ComponentInfoImpl[fullNames.length];
      for (int i = 0; i < comps.length; i++) {
        comps[i] = ComponentInfoImpl.fromString(fullNames[i]);
      }
    }
    return comps;
  }

  /**
   * Obtains the specified components (or their description only) from the specified SOFA node.
   * It may obtain more than just the explicitly asked components.
   * @param offers full names (the name and the implementation version in the format "name[version]")
   * of components whose description is to be obtained.
   * @param components full names (the name and the implementation version in the format "name[version]")
   * of components to be obtained (not only the description).
   * @param sofaNode a string representation of a reference to an InOut2InOut connector of a SOFA node.
   * It must not be <code>null</code>, otherwise an exception is thrown.
   * @param inferiors if <code>true</code>, all sub-components are obtained as well; otherwise, the
   * explicitly asked components are obtained only.
   * @throws InOutException if an error when connecting to the SOFA node occurs.
   */
  private void _obtain (String[] offers, String[] components, String sofaNode, boolean inferiors) throws InOutException {
    if (sofaNode == null) {
      throw new InOutException("The SOFA node where components are to be distributed from must not be null!");
    }

    InOut2InOut inOut = getConnection(sofaNode);

    ComponentInfoImpl[] descs = fullNamesToComponentInfos(offers);
    ComponentInfoImpl[] comps = fullNamesToComponentInfos(components);

    Bundle bundle = inOut.pullBundle(descs, comps, inferiors);
    tr.storeBundle(bundle);
  }

  /**
   * Distributes given components (or their description only) to given SOFA nodes.
   * It may distribute the explicitly specified components only as well as all sub-components
   * of the specified ones.
   * @param offers full names (the name and the implementation version in the format "name[version]")
   * of components whose description is to be distributed.
   * @param components full names (the name and the implementation version in the format "name[version]")
   * of components to be distributed (not only the description).
   * @param sofaNodes string representations of references to InOut2InOut connectors of SOFA nodes.
   * It must not be <code>null</code> nor an empty array, otherwise an exception is thrown.
   * @param inferiors if <code>true</code>, all sub-components are obtained as well; otherwise, the
   * explicitly asked components are obtained only.
   * @throws InOutException if an error occurs or the array of SOFA nodes is empty or <code>null</code>.
   */
  private void _distribute (String[] offers, String[] components, String[] sofaNodes, boolean inferiors) throws InOutException {

    if (sofaNodes == null || sofaNodes.length == 0) {
      throw new InOutException("The list of SOFA nodes where components are to be distributed to must not be null nor empty!");
    }

    ComponentInfoImpl[] descs = fullNamesToComponentInfos(offers);
    ComponentInfoImpl[] comps = fullNamesToComponentInfos(components);

    Bundle bundle = tr.getBundle(descs, comps, inferiors);
    for (int i = 0; i < sofaNodes.length; i++) {
      String sofaNode = sofaNodes[i];
      distributeComponent(bundle, sofaNode);
    }
  }

  /**
   * Obtains the specified components (or their description only) from the specified SOFA node.
   * It obtains the explicitly asked components only.
   * @param offers full names (the name and the implementation version in the format "name[version]")
   * of components whose description is to be obtained.
   * @param components full names (the name and the implementation version in the format "name[version]")
   * of components to be obtained (not only the description).
   * @param sofaNode a string representation of a reference to an InOut2InOut connector of a SOFA node.
   * It must not be <code>null</code>, otherwise an exception is thrown.
   * @throws InOutException if an error when connecting to the SOFA node occurs.
   */
  public void obtain (String[] offers, String[] components, String sofaNode) throws InOutException {
    _obtain(offers, components, sofaNode, false);
  }

  /**
   * Distributes given components (or their description only) to given SOFA nodes.
   * It distributes the explicitly specified components only.
   * @param offers full names (the name and the implementation version in the format "name[version]")
   * of components whose description is to be distributed.
   * @param components full names (the name and the implementation version in the format "name[version]")
   * of components to be distributed (not only the description).
   * @param sofaNodes string representations of references to InOut2InOut connectors of SOFA nodes.
   * It must not be <code>null</code> nor an empty array, otherwise an exception is thrown.
   * @throws InOutException if an error occurs or the array of SOFA nodes is empty or <code>null</code>.
   */
  public void distribute (String[] offers, String[] components, String[] sofaNodes) throws InOutException {
    _distribute(offers, components, sofaNodes, false);
  }

  /**
   * Obtains descriptions of the specified components and all their sub-components from the specified SOFA node
   * @param offers full names (the name and the implementation version in the format "name[version]")
   * of components whose description is to be obtained.
   * @param sofaNode a string representation of a reference to an InOut2InOut connector of a SOFA node.
   * It must not be <code>null</code>, otherwise an exception is thrown.
   * @throws InOutException if an error when connecting to the SOFA node occurs.
   */
  public void obtainOffers (String[] offers, String sofaNode) throws InOutException {
    _obtain(offers, null, sofaNode, true);
  }

  /**
   * Distributes descriptions of the given components and all their sub-components to the given SOFA nodes.
   * @param offers full names (the name and the implementation version in the format "name[version]")
   * of components whose description is to be distributed.
   * @param sofaNodes string representations of references to InOut2InOut connectors of SOFA nodes.
   * It must not be <code>null</code> nor an empty array, otherwise an exception is thrown.
   * @throws InOutException if an error occurs or the array of SOFA nodes is empty or <code>null</code>.
   */
  public void distributeOffers (String[] offers, String[] sofaNodes) throws InOutException {
    _distribute(offers, null, sofaNodes, true);
  }

  /**
   * Obtains the specified components and all their sub-components from the specified SOFA node.
   * @param components full names (the name and the implementation version in the format "name[version]")
   * of components to be obtained (not only the description).
   * @param sofaNode a string representation of a reference to an InOut2InOut connector of a SOFA node.
   * It must not be <code>null</code>, otherwise an exception is thrown.
   * @throws InOutException if an error when connecting to the SOFA node occurs.
   */
  public void obtainComponents (String[] components, String sofaNode) throws InOutException {
    _obtain(null, components, sofaNode, true);
  }

  /**
   * Distributes given components and all their sub-components to the given SOFA nodes.
   * @param components full names (the name and the implementation version in the format "name[version]")
   * of components to be distributed (not only the description).
   * @param sofaNodes string representations of references to InOut2InOut connectors of SOFA nodes.
   * It must not be <code>null</code> nor an empty array, otherwise an exception is thrown.
   * @throws InOutException if an error occurs or the array of SOFA nodes is empty or <code>null</code>.
   */
  public void distributeComponents (String[] components, String[] sofaNodes) throws InOutException {
    _distribute(null, components, sofaNodes, true);
  }

  /**
   * Gets a bundle containing a desription of all components installed on this SOFA node.
   * @return a bundle decribing all components.
   */
  public Bundle list () {
    return tr.list();
  }

}
