/* 
 * $Id: CompositionNode.java,v 1.4 2005/07/08 12:04:12 kofron Exp $ 
 *
 * Copyright 2004
 *    Distributed Systems Research Group
 *    Department of Software Engineering
 *    Faculty of Mathematics and Physics
 *    Charles University, Prague
 *
 * Copyright 2005
 *    Formal Methods In Software Engineering Group
 *    Institute of Computer Science
 *    Academy of Sciences of the Czech Republic
 * 
 * This code was developed by Jan Kofron <kofron@nenya.ms.mff.cuni.cz>
 */

package SOFA.SOFAnode.Util.DFSRChecker.node;

import java.util.TreeSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;

import SOFA.SOFAnode.Util.DFSRChecker.DFSR.CheckingException;
import SOFA.SOFAnode.Util.DFSRChecker.state.CompositeState;
import SOFA.SOFAnode.Util.DFSRChecker.state.State;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPair;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPairs;
import SOFA.SOFAnode.Util.DFSRChecker.utils.AnotatedProtocol;


/**
 * Represents the old composition operator.
 */
public class CompositionNode extends TreeNode {

    /** Creates a new instance of CompositionNode */
    public CompositionNode(TreeNode node1, TreeNode node2, TreeSet events, ActionRepository repository) {
        //super(getUnion(node1, node2));
        super("(" + node1.protocol + ")&(" + node2.protocol + ")");
        this.events = events;
        this.repository = repository;

        this.nodes = new TreeNode[2];
        this.nodes[0] = node1;
        this.nodes[1] = node2;
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getInitial()
     */
    public State getInitial() {
        State[] states = new State[2];
        states[0] = nodes[0].getInitial();
        states[1] = nodes[1].getInitial();

        return new CompositeState(states);
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#isAccepting(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public boolean isAccepting(State state) {
        CompositeState cstate = (CompositeState) state;

        return (nodes[0].isAccepting(cstate.states[0])) && (nodes[1].isAccepting(cstate.states[1]));
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getTransitions(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public TransitionPairs getTransitions(State state) throws InvalidParameterException, CheckingException {
        CompositeState cstate = (CompositeState) state;

        TransitionPair[] trans1 = nodes[0].getTransitions(cstate.states[0]).transitions;
        TransitionPair[] trans2 = nodes[1].getTransitions(cstate.states[1]).transitions;

        ArrayList result = new ArrayList();
        TreeMap sync = new TreeMap();

        //we divide the transitions of node1 (states[0]) into two groups
        // - that of not present within the events set and the others
        for (int i = 0; i < trans1.length; ++i) {
            Integer key = new Integer(trans1[i].eventIndex);

            if (events.contains(new Integer(ActionRepository.getPure(trans1[i].eventIndex))))
                if (sync.containsKey(key))
                    ((ArrayList) sync.get(key)).add(trans1[i]);
                else {
                    ArrayList value = new ArrayList();
                    value.add(trans1[i]);
                    sync.put(new Integer(trans1[i].eventIndex), value);
                }
            else {
                State[] states = new State[2];
                states[0] = trans1[i].state;
                states[1] = cstate.states[1];

                result.add(new TransitionPair(trans1[i].eventIndex, new CompositeState(states)));
            }
        }

        // now complete the result array
        for (int i = 0; i < trans2.length; ++i) {
            Integer key = new Integer(trans2[i].eventIndex);
            if (events.contains(new Integer(ActionRepository.getPure(trans2[i].eventIndex)))) {
                Integer invkey = new Integer(repository.getInverseAction(trans2[i].eventIndex));
                if (sync.containsKey(invkey)) {
                    Iterator it = ((ArrayList) sync.get(invkey)).iterator();

                    while (it.hasNext()) {
                        State[] states = new State[2];
                        states[0] = (State) it.next();
                        states[1] = trans2[i].state;

                        result.add(new TransitionPair(repository.getTauAction(trans2[i].eventIndex), new CompositeState(states)));
                    }
                }
            } else {
                State[] states = new State[2];
                states[0] = cstate.states[0];
                states[1] = trans2[i].state;

                result.add(new TransitionPair(trans2[i].eventIndex, new CompositeState(states)));
            }

        }

        TransitionPair[] result1 = new TransitionPair[result.size()];
        Iterator it = result.iterator();
        int i = 0;

        while (it.hasNext())
            result1[i++] = (TransitionPair) it.next();

        return new TransitionPairs(result1);

    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getWeight()
     */
    public long getWeight() {
        if (weight != -1)
            return weight;
        else {
            return weight = nodes[0].getWeight() * nodes[1].getWeight();
        }
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#forwardCut(java.util.TreeSet)
     */
    public TreeNode forwardCut(TreeSet livingevents) {
        TreeSet exlivev = new TreeSet(livingevents);

        exlivev.addAll(events);
        TreeSet exlivev2 = new TreeSet(exlivev);

        nodes[0] = nodes[0].forwardCut(exlivev);
        nodes[1] = nodes[1].forwardCut(exlivev2);

        if ((nodes[0] == null) || (nodes[1] == null))
            return null;

        else
            return this;
    }

    /**
     * @return the symbolic name of the treenode denoting its type
     */
    public String[] getTypeName() {
        String[] result = { "Composition", "&" };
        return result;
    }


    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getAnotatedProtocol(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public AnotatedProtocol getAnotatedProtocol(State state) {
        throw new RuntimeException(this.getClass().getName() + ".getAnotatedProtocol is not implemented");
    }


    
    //--------------------------------------------------------------------------

    /**
     * The set of event for composition operator
     */
    final private TreeSet events;

    /**
     * Action repository for seeking the tau and other event indices
     */
    private ActionRepository repository;

}