/* 
 * $Id: DeterministicNode.java,v 1.4 2005/07/08 12:04:12 kofron Exp $ 
 *
 * Copyright 2004
 *    Distributed Systems Research Group
 *    Department of Software Engineering
 *    Faculty of Mathematics and Physics
 *    Charles University, Prague
 *
 * Copyright 2005
 *    Formal Methods In Software Engineering Group
 *    Institute of Computer Science
 *    Academy of Sciences of the Czech Republic
 * 
 * This code was developed by Jan Kofron <kofron@nenya.ms.mff.cuni.cz>
 */

package SOFA.SOFAnode.Util.DFSRChecker.node;


import java.util.ArrayList;
import java.util.LinkedList;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import SOFA.SOFAnode.Util.DFSRChecker.DFSR.CheckingException;
import SOFA.SOFAnode.Util.DFSRChecker.state.CompositeState;
import SOFA.SOFAnode.Util.DFSRChecker.state.State;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPair;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPairs;
import SOFA.SOFAnode.Util.DFSRChecker.utils.AnotatedProtocol;

/**
 * Represented operation of determinization.
 */
public class DeterministicNode extends TreeNode {

    /**
     * Creates a new instance of DeterministicNode
     */
    public DeterministicNode(TreeNode node) {
        super("Det(" + node.protocol + ")");
        this.nodes = new TreeNode[1];
        this.nodes[0] = node;
    }

    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getInitial()
     */
    public State getInitial() {
        State[] state = new State[1];
        state[0] = nodes[0].getInitial();

        return new CompositeState(state);
    }

    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#isAccepting(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public boolean isAccepting(State state) {
        CompositeState cstate = (CompositeState) state;

        for (int i = 0; i < cstate.states.length; ++i)
            if (nodes[0].isAccepting(cstate.states[i]))
                return true;

        return false;

    }

    /**
     * This method creates the deterministic transition set form a
     * nedeterministic one. It creates, of course, for each of the possible
     * transitions via the same event a transition to a newly created
     * CompositeState.
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getTransitions(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public TransitionPairs getTransitions(State state) throws InvalidParameterException, CheckingException {
        CompositeState cstate = (CompositeState) state;
        TreeMap transmap = new TreeMap();
        LinkedList deps = new LinkedList();

        int nodecnt = cstate.states.length;

        for (int i = 0; i < nodecnt; ++i) {
            TransitionPairs transp = nodes[0].getTransitions(cstate.states[i]);
            TransitionPair[] trans = transp.transitions;
            
            if (transp.deps != null)
                deps.addAll(transp.deps);
            
            for (int j = 0; j < trans.length; ++j) {
                TreeSet states = (TreeSet) transmap.get(new Integer(trans[j].eventIndex));
                if (states != null) {
                    states.add(trans[j].state);
                } else {
                    states = new TreeSet();
                    states.add(trans[j].state);
                    transmap.put(new Integer(trans[j].eventIndex), states);
                }
            }
        }

        TransitionPair[] result = new TransitionPair[transmap.size()];
        Set keys = transmap.keySet();
        Iterator it = keys.iterator();
        int i = 0;

        while (it.hasNext()) {
            Integer eventindex = (Integer) it.next();
            TreeSet states = (TreeSet) transmap.get(eventindex);
            State[] resultstates = new State[states.size()];

            Iterator it2 = states.iterator();
            int l = 0;

            while (it2.hasNext())
                resultstates[l++] = (State) it2.next();

            result[i++] = new TransitionPair(eventindex.intValue(), new CompositeState(resultstates));
        }

        //TODO: debug only!
        /*
        for (i = 0; i < result.length; ++i)
            result[i].state.getSignature();
        */
        
        TransitionPairs restrans = new TransitionPairs(result);
        restrans.deps = deps;
        return restrans;
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getWeight()
     */
    public long getWeight() {
        if (weight != -1)
            return weight;
        else {
            return weight = nodes[0].getWeight();
        }
    }
    
    /**
     * @return the symbolic name of the treenode denoting its type
     */
    public String[] getTypeName() {
        String[] result = { "Deterministic", "DET" };
        return result;
    }
    
    

    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getAnotatedProtocol(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public AnotatedProtocol getAnotatedProtocol(State state) {
        CompositeState cstate = (CompositeState) state;
        String result = new String();
        int cnt = cstate != null ? cstate.states.length : 0;
        ArrayList indicesresult = new ArrayList();

        for (int i = 0; i < cnt; ++i) {
            AnotatedProtocol subresult = nodes[0].getAnotatedProtocol(cstate.states[i]);
            
            for (int j = 0; j < subresult.indices.size(); ++j)
                indicesresult.add(new Integer(((Integer)subresult.indices.get(j)).intValue() + result.length()));
            
        }
        
        result = nodes[0].getAnotatedProtocol(null).protocol;
        
        return new AnotatedProtocol(result, indicesresult);
    }
    

    //---------------------------------------------------------------
}