/* 
 * $Id: TreeNode.java,v 1.4 2005/07/08 12:04:12 kofron Exp $ 
 * 
 * Copyright 2004
 *    Distributed Systems Research Group
 *    Department of Software Engineering
 *    Faculty of Mathematics and Physics
 *    Charles University, Prague
 *
 * Copyright 2005
 *    Formal Methods In Software Engineering Group
 *    Institute of Computer Science
 *    Academy of Sciences of the Czech Republic
 * 
 * This code was developed by Jan Kofron <kofron@nenya.ms.mff.cuni.cz>
 */

package SOFA.SOFAnode.Util.DFSRChecker.node;

import java.util.TreeSet;

import SOFA.SOFAnode.Util.DFSRChecker.DFSR.CheckingException;
import SOFA.SOFAnode.Util.DFSRChecker.state.State;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPairs;
import SOFA.SOFAnode.Util.DFSRChecker.utils.AnotatedProtocol;


/**
 * This class represents the node of a tree describing the state space of an
 * "execution".
 */
abstract public class TreeNode implements Cloneable {

    /** Creates a new instance of TreeNode */
    public TreeNode(String protocol) {
        this.protocol = protocol;
        weight = -1;
    }

    /**
     * @return the expected weight of this node, i.e. the size of automaton
     *         generated by its subtree.
     */
    abstract public long getWeight();

    /**
     * @return the vector of node's children.
     */
    public TreeNode[] getChildren() {
        return nodes;
    }

    /**
     * Changes the pointer at position of index to the newChild.
     */
    public void changeChild(int index, TreeNode newChild) throws ArrayIndexOutOfBoundsException {
        nodes[index] = newChild;
    }

    /**
     * Performs the forward cutting optimalization for this node. This is the
     * default implementation for the majority of node types
     * 
     * @param livingevents
     *            the already active events
     * @return the "optimized" tree, i.e. without some nodes and edges
     */
    public TreeNode forwardCut(TreeSet livingevents) {
        TreeNode active = null;
        int cnt = 0;

        for (int i = 0; i < nodes.length; ++i) {
            nodes[i] = nodes[i].forwardCut(livingevents);
            if (nodes[i] != null) {
                ++cnt;
                active = nodes[i];
            }
        }

        // no child node's survived
        if (cnt == 0)
            return null;

        // only one child node's survived and it hasn't been the only child
        // else we have to return the original node
        else if ((cnt == 1) && (nodes.length > 1))
            return active;

        // some of the child nodes have died
        else if (nodes.length != cnt) {
            TreeNode[] newchildren = new TreeNode[cnt];

            for (int i = 0, j = 0; i < cnt; ++i) {
                while (nodes[j] == null)
                    ++j;

                newchildren[i] = nodes[j++];
            }

            nodes = newchildren;

        }

        return this;
    }

    /**
     * This method creates a new (deep) copy of this node. This is needed for forward
     * cutting optimization.
     * 
     * @return the deepcopy of this node
     */
    public TreeNode copy() {
        TreeNode newnode = null;
        try {
            newnode = (TreeNode) this.clone();
        } catch (CloneNotSupportedException e) {
            return null;
        }

        newnode.nodes = new TreeNode[this.nodes.length];

        for (int i = 0; i < this.nodes.length; ++i) {
            newnode.nodes[i] = this.nodes[i].copy();
        }

        return newnode;
    }
    
    /**
     * Counts the number of leaves of this subtree
     * @return	the number of leaves
     */
    public int getLeafCount() {
        int cnt = 0;
        for (int i = 0; i < nodes.length; ++i)
            cnt += nodes[i].getLeafCount();
        
        return cnt;
    }
    
    /**
     * @return AnotatedProtocol for printing and error trace understanding.
     */
    abstract public AnotatedProtocol getAnotatedProtocol(State state);

    //--------------------------------------------------------------------------
    /**
     * @return the initial state of the automat
     */
    abstract public State getInitial();

    /**
     * @return true if this is an accepting state, false otherwise
     */
    abstract public boolean isAccepting(State state);

    /**
     * Retrieve the list of transition from a given state
     * 
     * @param state find all transitions from this state
     * @return the list of pairs of transitions (and their names)
     * @throws InvalidParameterException	means internal error - bad parameter passed to an eventnode
     * @throws CheckingException		means a composition error was detected
     */
    abstract public TransitionPairs getTransitions(State state) throws InvalidParameterException, CheckingException;

    /**
     * @return the symbolic name of the treenode denoting its type
     */
    abstract public String[] getTypeName();

    //--------------------------------------------------------------------------
    /** private attributes */

    /**
     * Child nodes of this node
     */
    protected TreeNode[] nodes;

    /**
     * The weight of the node - it is counted in a lazy way
     */
    protected long weight;

    /**
     * The protocol represented by this node
     */
    public String protocol;

}

