/**
 * Tools.java is a part of the SOFA project.
 * This file was created by pepan on 3.5.2003.
 */
package SOFA.Util;

import java.io.*;

/**
 * Contains a few useful commonly used tools.
 * @author  Petr Panuska and Ladislav Sobr
 */
public class Tools {
  /**
   * Deletes the given file. In case the file is a directory, it deletes all
   * sub-directories/files lying under this directory.
   * @param file the root directory or file which is going to be deleted.
   */
  public static void deleteSubTree (File file) {
    if (file.isDirectory()) {
      File[] subFiles = file.listFiles();
      for (int i = 0; i < subFiles.length; i++) {
        File subFile = subFiles[i];
        Tools.deleteSubTree(subFile); // recursively delete sub-files first
      }
    }
    file.delete();
  }
  
  /**
   * If the 'file' is empty directory, it is deleted. If the parent of the 'file'
   * becomes empty directory, it is also deleted - and continue recursively.
   * This deletion stops either on non-empty directory or on 'stopDir'
   */
  
  public static void deleteEmptyDirectoryTree(File file, File stopDir)
  {
    while (file != null && file.isDirectory() && !file.equals(stopDir) && file.list().length == 0)
    {
      file.delete();
      file = file.getParentFile();
    }
  }

  
  /**
   * Moves a file. First it tries to rename the file. If it fails, it simply copies
   * the content of the source file to the destination file and deletes the source file.
   * It keeps original time of last modification even in case of copying files.
   * @param srcFile the source file name (what is to be moved).
   * @param destFile the destination file (where it is to be moved).
   */
  public static void moveFile (File srcFile, File destFile) throws IOException
  {
    destFile.getParentFile().mkdirs();
    if (srcFile.renameTo(destFile) == false) { // try use the renaming method first
      // if failed, copy the file
      FileInputStream src = new FileInputStream(srcFile);
      FileOutputStream dest = new FileOutputStream(destFile);

      copyStream(src, dest);

      src.close();
      dest.close();
      destFile.setLastModified(srcFile.lastModified()); // keep the original time of creation
      srcFile.delete();
    }
  }
  
  /**
   * The length of a buffer used for reading data from a stream.
   */
  private static final int BUFFER_SIZE = 4096;

  /**
   * The buffer used for reading data from a stream.
   */
  private static final byte[] buffer = new byte[BUFFER_SIZE];

  /**
   * It copies all data from one stream to the second one using a static buffer.
   * Since the access to this buffer is synchronized, this method can be used
   * by many threads at one time with no problems.
   * @param src the input stream the data is read from.
   * @param dest the output stream the data is write to.
   * @throws IOException
   */
  public static void copyStream (InputStream src, OutputStream dest) throws IOException {
    int length;
    for (; ;) {
      synchronized (buffer) {
        length = src.read(buffer);
        if (length > 0)
          dest.write(buffer, 0, length);
        else
          break;
      }
    }
  }

}
